/****************************************************************************
 * This file is part of the program alifritz                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <cmath>
#include <cfloat>
#include <cstdio>

using namespace std;

#include "seqcode.hh"
#include "parser.hh"
#include "eingabe.hh"

#define MAXZEILENLAENGE  1000

#define DEFSUBSTMODEL JC69
#define DEFINDMODEL TKF1
#define DEFTREEFILE "treefile"
#define DEFALIFILE "alifile"
#define DEFSEQFILE "seqfile"

eingabe_parameter::eingabe_parameter()
{
}

void eingabe_parameter::from_file(char * dateiname)
    /*
     * Eingabe der eingabe_parameter aus Datei dateiname 
     */
{
    scriptpart alt=scNONE;
    scriptpart aktuell=scNONE;
    subst_option so;
    indel_option io;
    files_option fo;
    option_arg oa;
    string argument, Zeile;
    char *cZeile;
    
    ifstream inputstr(dateiname);
    if (!inputstr) {
	cerr << "cannot open file " << dateiname << endl;
	exit(EXIT_FAILURE);
    }
    if((cZeile=(char *)malloc(MAXZEILENLAENGE))==NULL) {
       cerr << "Cannot allocate buffer for the control file.\n";
       exit(EXIT_FAILURE);
    }
    while(!inputstr.eof()) {
        inputstr.getline(cZeile,MAXZEILENLAENGE);
	Zeile = string(cZeile);
        aktuell = isnewpart(Zeile);
        if(alt!=aktuell && aktuell!=scNONE)
	   alt=aktuell;
	else {
           switch(alt) {
               case scNONE :
	          break;
	       case SUBST :
	          so = whichsubstoption(Zeile,argument,oa);
	          if(so!=suNONE)
		     sub.adjust(so,argument,oa);
	          break;
	       case INDEL :
	          io = whichindeloption(Zeile,argument,oa);
	          if(io!=inNONE)
		     ind.adjust(io,argument,oa);
	          break;
               case FILES :
	          fo = whichfilesoption(Zeile,argument,oa);
	          if(fo!=fiNONE)
	             fil.adjust(fo,argument);
                  break;
               default :
	          cerr << "Unknown script part\n";
		  exit(EXIT_FAILURE);
           }
       }
    }
    inputstr.close();
    
    sub.abgleich();  // passt auch alles?
    ind.abgleich();
    
}



// substmod_parameters enthaelt die Einstellungen,
// die man fuer das Substitutionsmodell gewaehlt hat.
// model = Typ des Substitutionsmodells
// parameter = Parameter des Modells
// freqs = Zustandshaeufigkeiten
// freqset = sollen Basenhaeufigkeiten geschaetzt werden 
// Die Klasse substmod_parameters wird benutzt, um das in
// der Klasse mult_alignment enthaltene Substitutionsmodell
// zu konstruieren. 

substmod_parameters::substmod_parameters() // default constructor
{
   model = DEFSUBSTMODEL;
   modelset = FALSE;
   paramset = FALSE;
   freqset = FALSE;
}

void substmod_parameters::adjust(subst_option so,string argument,option_arg oa)
{
   switch(so) {
       case SMODEL :
            setmodel(argument,oa);
	    break;
       case SPARAM :
            setparams(argument,oa);
	    break;
       case FREQS :
            setfreqs(argument,oa);
	    break;
       default :
           cerr << "Impossible option "<<(int)so<<" under headline SUBST\n";
	   exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setmodel(string argument,option_arg oa)
{

   if(modelset) {
      cerr << "\nWARNING:\n"
           << "The substitution model has already been set.\n"
           << "The older value is now overwritten.\n\n";
   }
   modelset = TRUE;
   switch(oa) {
       case VALUE :
            model = stringtosubarch(argument);
            break;
       case DEFAULT :
	    model = DEFSUBSTMODEL;
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option model\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setparams(string argument,option_arg oa)
{
   if(paramset) {
      cerr << "\nWARNING:\n"
           << "The substitution model's parameters have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   switch(oa){
       case DEFAULT :
            paramset = FALSE;
            break;
       case VECTOR :
            paramset = TRUE;
            stringtovector(argument,params);
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option parameter\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setfreqs(string argument,option_arg oa)
{
   if(freqset) {
      cerr << "\nWARNING:\n"
           << "The base or amino acid frequencies have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   switch(oa){
       case DEFAULT :
            freqset = FALSE;
            break;
       case VECTOR :
            freqset = TRUE;
            stringtovector(argument,freqs);
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option freqs\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::abgleich()
{
   unsigned int i;
   bool flag;
   bool viertelflag;
   
   if(modelset) {
     
     // Welche Wahl wurde fuer die Subst.raten getroffen?
     switch(model) {
          case FEL81 :
          case JC69 :
             if(paramset && params.size()>0) {
                cerr << "The Felsenstein model and the Jukes Cantor do not use\n"
	             << "any rate parameters.\n";
		exit(EXIT_FAILURE);
	     }
	     params.clear();
             break;
          case HKY :
          case KIMURA2P :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=1) {
                    cerr << "The HKY and the K2P model have only one rate parameter.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    if(params[0]<=0.0) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear(); 
                  params.push_back(0.5);
	     }
	     break;
          case GENREV :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=6) {
                    cerr << "The general reversible model has six rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
		  for(i=0;i<6;i++) 
                    params.push_back(1.0);
	     }
	     break;
          case TAMNEI :
          case KIMURA3P :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=2) {
                    cerr << "The Tamura-Nei and the Kimura3P model have two rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
                  params.push_back(0.5);
                  params.push_back(1.0);
	     }
             break;
          case MV2000 :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=190) {
                    cerr << "The general reversible model for proteins has 190 rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
                  for(i=0;i<190;i++)
                    params.push_back(1.0);
	     }
             break;
          case DAYHOFF :
          case BLOSUM62 :
          case JTT :
          case MTREV24 :
	     if(paramset || freqset) {
	        cerr << "The substitution model that you want to use does not\n"
		     << "accept user specified rate parameters or amino acid\n"
		     << "frequencies.\n";
		exit(EXIT_FAILURE);
	     }
             break;   
          default :
	     cerr << "Unknown substitution model\n";
	     exit(EXIT_FAILURE);
     }
     
     // Welche Wahl wurde fuer die Zustandshaeufigkeiten getroffen?
     if(freqset) { // Wenn nichts angegeben wurde, wird eh geschaetzt
         flag=TRUE;
	 for(i=0;i<freqs.size();i++)
	     flag = flag && (freqs[i]>0.0);
	 if(!flag) {
	    cerr << "All nucleotide or amino acid frequencies have to be positive.\n";
	    exit(EXIT_FAILURE);
	 }
	 // Ob die Summe der Haeufigkeiten 1 ist, wird nicht ueberprueft,
	 // da der Konstruktor von substmod mit absoluten Haeufigkeiten
	 // umgehen kann
         switch(model) {
              case JC69 :
              case KIMURA2P :
              case KIMURA3P :
              case DAYHOFF :
              case BLOSUM62 :
              case JTT :
              case MTREV24 :
	         cerr << "The substitution model that you want to use does not\n"
		      << "accept user specified rate parameters or state frequencies.\n";
		 exit(EXIT_FAILURE);
                 break; 
              case FEL81 :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the FEL81 model becomes the JC69 model.\n";
		    model = JC69;
		 }
                 break;
              case HKY :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the HKY model becomes the KIMURA2P model.\n";
		    model = KIMURA2P;
		 }
	         break;
              case GENREV :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 break;
              case TAMNEI :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the Tamura-Nei model becomes the KIMURA3P model.\n";
		    model = KIMURA3P;
		 }
	         break;
              case MV2000 :
	         if(freqs.size()!=20) {
		    cerr << "Please specify exactly twenty amino acid frequencies.\n";
		    exit(EXIT_FAILURE);
		 }
		 break;
              default :
	         cerr << "Unknown substitution model\n";
	         exit(EXIT_FAILURE);
         }
     }

   } else {
       if(paramset) {
          cerr << "Please specify a substitution model using the option 'model=X'\n"
	       << "Possible choices for X are:\n"
	       << "for nucleotides:             for amino acids:\n"
	       << "   FEL81                        MV2000\n"
	       << "   JC69                         DAYHOFF\n"
	       << "   HKY                          BLOSUM62\n"
	       << "   KIMURA2P                     JTT\n"
	       << "   TAMNEI                       MTREV24\n"
	       << "   KIMURA3P\n"
	       << "   GENREV\n\n";
	  exit(EXIT_FAILURE);
       } else {
           if(freqset) {
              cerr << "Please specify a substitution model using the option 'model=X'\n"
	           << "Possible choices for X are:\n"
	           << "for nucleotides:             for amino acids:\n"
	           << "   FEL81                        MV2000\n"
	           << "   JC69                         DAYHOFF\n"
	           << "   HKY                          BLOSUM62\n"
	           << "   KIMURA2P                     JTT\n"
	           << "   TAMNEI                       MTREV24\n"
	           << "   KIMURA3P\n"
	           << "   GENREV\n\n";
	       exit(EXIT_FAILURE);	   
	   } 
	   else
	       model=JC69;
       }
   }
}

alphsize substmod_parameters::retalf()
{
     switch(model) {
         case FEL81:
	 case JC69:
	 case HKY:
	 case KIMURA2P:
	 case TAMNEI:
	 case KIMURA3P:
	 case GENREV:
	    return DNA;
	 case MV2000:
	 case DAYHOFF:
	 case BLOSUM62:
	 case JTT:
	 case MTREV24:
            return PROTEIN;
	 default:
	    cerr << "unknown substitution model\n";
	    exit(EXIT_FAILURE);
     }
     
     return DOUBLET;
}

vector<double>& substmod_parameters::retfreqs()
{
     if(!freqset) { //keine Haeufigkeiten angegeben
         int alf = retalf();
	 freqs.resize(alf);
	 for(unsigned int i=0;i<freqs.size();++i)
	     freqs[i]=1.0;  // absolute Haeufigkeiten reichen
     }

     return freqs;
}

// indel_parameters enthaelt die Einstellungen fuer das Insertions-Deletionsmodell.
// Analog zu substmod_parameters
indel_parameters::indel_parameters()
{
   model = DEFINDMODEL;
   modelset = FALSE;
   paramset = FALSE;
   switch(model) {
     case TKF1 :
       params.push_back(0.01);
       params.push_back(0.011);
       break;
     case TKF2 :
       params.push_back(0.01);
       params.push_back(0.011);
       params.push_back(0.5);   // Parameter der Fragmentlaenge
       break;
     default :
       cerr << "At the moment there is only TKF1 and TKF2\n";
       exit(EXIT_FAILURE);
   }
}

void indel_parameters::adjust(indel_option io,string argument,option_arg oa)
{
   switch(io) {
       case IMODEL :
            setmodel(argument,oa);
	    break;
       case IPARAM :
            setparams(argument,oa);
	    break;
       default :
           cerr << "Impossible option under headline INDEL\n";
	   exit(EXIT_FAILURE);
   }
}

void indel_parameters::setmodel(string argument,option_arg oa)
{

   if(modelset) {
      cerr << "\nWARNING:\n"
           << "The insertion deletion model has already been set.\n"
           << "The older value is now overwritten.\n\n";
   }
   modelset = TRUE;
   switch(oa) {
       case VALUE :
            model = stringtoidarch(argument);
            break;
       case DEFAULT :
	    model = DEFINDMODEL;
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option model\n";
	    exit(EXIT_FAILURE);
   }
}

void indel_parameters::setparams(string argument,option_arg oa)
{
   if(paramset) {
      cerr << "\nWARNING:\n"
           << "The insertion deletion model's parameters have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   double d;
   switch(oa){
       case DEFAULT :
            paramset = FALSE;
            break;
       case VECTOR :
            paramset = TRUE;
            stringtovector(argument,params);
// workaround (not nice)
            if(params.empty()){
              cerr<<"Please specify parameters for the indel model or\n"
                  <<"remove the line 'parameter=' from the control file.\n";
              exit(EXIT_FAILURE);
            }
            d=params.back();
            params.push_back(d);
            params[1]=params[0]+0.001;
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option parameter\n";
	    exit(EXIT_FAILURE);
   }
}

void indel_parameters::abgleich()
{
    if(modelset) {
      switch(model) {
         case TKF1 :
	    if(paramset) {
	      if(params.size()!=2) {
	         cerr << "The TKF1 model has two parameters.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[0]<=0.0 || params[1]<=0.0 || params[1]<=params[0]) {
	         cerr << "Both rates of the TKF1 model have to be positive and\n"
		      << "the first rate has to be smaller than the second one.\n";
		 exit(EXIT_FAILURE);
	      }
	    } else {
	      params.clear();
              params.push_back(0.01);
              params.push_back(0.011);
	    }
	    break;
         case TKF2 :
	    if(paramset) {
	      if(params.size()!=3) {
	         cerr << "The TKF2 model has three parameters.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[0]<=0.0 || params[1]<=0.0 || params[1]<=params[0]) {
	         cerr << "Both rates of the TKF2 model have to be positive and\n"
		      << "the first rate has to be smaller than the second one.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[2]<=0.0 || params[2]>=1.0) {
	         cerr << "The third parameter of the TKF2 model should be positive\n"
		      << "and smaller than 1.\n";
		 exit(EXIT_FAILURE);
	      }
	    } else {
	      params.clear();
              params.push_back(0.01);
              params.push_back(0.011);
              params.push_back(0.5);
	    }
	    break;
         default :
	    cerr << "Unknown insertion and deletion model\n";
	    exit(EXIT_FAILURE);
      }
    } else {
      if(paramset) {
        cerr << "Please specify an insertion and deletion model.\n"
	     << "Possible choices are:\n"
	     << "     TKF1\n"
	     << "     TKF2\n\n";
	exit(EXIT_FAILURE);
      } else {
        model = TKF1;
	params.clear();
        params.push_back(0.01);
        params.push_back(0.011);
      }
    }
}

// filenames enthaelt die gewuenschten Dateinamen
filenames::filenames()
{
   treefile=DEFTREEFILE;
   alifile=DEFALIFILE; 
   seqfile=DEFSEQFILE; 
}

void filenames::adjust(files_option fo, string argument)
{
   switch(fo) {
      case TREEFILE :
         treefile=argument; //whichfilesoption sollte schon den richtigen Teil extrahiert haben
         break;
      case ALIFILE :
         alifile=argument;
         break;
      case SEQFILE :
         seqfile=argument;
         break;
      default :
         cerr << "Impossible option under headline FILES\n";
	 exit(EXIT_FAILURE);
   }
}

