/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#ifndef INDEL_H
#define INDEL_H

/************** enums zur konsistenten Indizierung *******************************/
enum zustand {PAIR=0,INSERTION=1,DELETION=2,START=3,ENDE=4,NOSTATE=5,SUSPEND=6};

//enum fbzustand {fbSTART=0,fbY1=1,fbXY=2,fbY2=3,fbX_=4,fbY3=5,fbENDE=6,fbNOSTATE=7};

//SUSPEND entspricht --
enum triplet_zus {tXYZ=0,tXY_=1,tX_Z=2,t__Z=3,t_Y_1=4,t_Y_2=5,
                  tX__=6,tSTART=7,tENDE=8,t_YZ=9,tNOSTATE=10};
/* Dargestellt wird der folgende 'Baum': 
	Y
       /
      X  
       \ 
        Z	   */
// links Vorfahr, Mitte linker Nachkomme, rechts rechter Nachkomme;
// t_Y_1 erreicht man von tSTART, tXYZ und tX_Z,
// t_Y_2 von tXY_ und tX__
// t_YZ ist nur als Aequivalent von q_XYZ von Bedeutung (sonst unmoeglich)

enum quart_zus {qWXYZ = 0, qWXY_ = 1, qWX_Z = 2, q_XYZ = 3, qWX__ = 4,
                q_XY_ = 5, q_X_Z = 6, qW___ = 7, q_X__ = 8, q__Y_1 = 9,
		q__Y_2=10, q__Y_3=11, q__Y_4=12, q___Z1=13, q___Z2=14,
		qSTART=15, qENDE=16,  qNOSTATE=17, qW_YZ = 18, qW_Y_ = 19,
		qW__Z = 20, q__YZ = 21};
/* Dargestellt wird der folgende Baum: 
	Y
       /
   W--X  
       \ 
        Z	   */
// links Grossvater, Mitte links Vater, Mitte rechts linker Sohn, rechts rechter Sohn;
// q__Y_1 erreicht man von qSTART, q_XYZ und q_X_Z,
// q__Y_2 von q_XY_ und q_X__,
// q__Y_3 von qWXYZ und qWX_Z,
// q__Y_4 von qWXY_ und qWX__;
// q___Z1 erreicht man von qSTART, q_XYZ, q_X_Z, q_XY_, q_X__, q__Y_1 und q__Y_2,
// q___Z2 von qWXYZ, qWX_Z, qWXY_, qWX__, q__Y_3 und q__Y_4.
// Die letzten vier Zustaende sind im 4er-Baum unmoeglich und dienen nur
// zur Beschreibung eines 4-Taxonalignments ohne innere Sequenz.

enum hex_zus {hUVWXYZ = 0, hUV_XYZ = 1, hUVWX_Z = 2, hUV_X_Z = 3,
              hUVWXY_ = 4, hUV_XY_ = 5, hUVWX__ = 6, hUV_X__ = 7,
	      hUVW___ = 8, hUV____ = 9, hU_____ = 10, h_VWXYZ = 11,
	      h_V_XYZ = 12, h_VWX_Z = 13, h_V_X_Z = 14, h_VWXY_ = 15,
	      h_V_XY_ = 16, h_VWX__ = 17, h_V_X__ = 18, h_VW___ = 19,
	      h_V____ = 20, hW1 = 21, hW2 = 22, hW3 = 23, hW4 = 24, hW5 = 25,
	      h___XYZ = 26, h___X_Z = 27, h___XY_ = 28, h___X__ = 29, hY1 = 30,
	      hY2 = 31, h_____Z = 32, hSTART = 33, hENDE = 34, hNOSTATE = 35};
/* Dargestellt wird der folgende Baum: 
          W
	 /
	/
   U---V   Y
        \ /
	 X
	  \
	   Z
*/


/*************** Indel-Modell **************************************************/
class fbindel;

class indel {
 friend class fbindel;
 private :
   vector<double> params; //enthaelt die Modellparameter
   idarch model;  // Modelltyp(TKF1=1)
   double pt, ty, tz, qw, qy, qz, huv, hvw, hvx, hxy, hxz; // Zeiten
   vector<vector<double> > P; //logUe'w'keiten fuer Sequenzpaar
   vector<vector<double> > T; //logUe'w'keiten fuer triplet_zus
   vector<vector<double> > Q; //logUe'w'keiten fuer quart_zus
   vector<vector<double> > H; //logUe'w'keiten fuer hex_zus
 public :
   indel(); //Default-Konstruktor
   indel(double lambda, double mu); //TKF1-Konstruktor;
   indel(vector<double> parameter, idarch  mod); //allg. Konstruktor
   idarch returnparams(vector<double> &v);  //gibt die gueltigen Parameterwerte zurueck
   void setparams(vector<double> &newpars); //setzt neue Parameterwerte
   void rescale(double scale); // neuskalierung der Parameter
   double tkf1_alpha(double mue, double time); 
   double tkf1_beta(double lambda, double mue, double time);
   double tkf1_gamma(double lambda, double mue, double time);
   double logprob_of_seqlen(unsigned int seqlen);
   void fillP(double time);
   void fillT(double txy, double txz);
   void fillQ(double twx, double txy, double txz);
   void fillH(double tuv, double tvw, double tvx, double txy, double txz);
   double retprob(zustand z1, zustand z2);
   double retlnprob(int z1, int z2);
   zustand nextpair(zustand now, bool stillseq);
   triplet_zus nexttriplet(triplet_zus now, zustand then, unsigned int internalcount);
   quart_zus nextquartet(quart_zus now, triplet_zus then, unsigned int internalcount);
   hex_zus nexthextet(hex_zus now, quart_zus then, unsigned int internalcount);
   indel operator=(indel id2);
   void setexplength(unsigned int explen); // veraendert mue so, dass explen die erwartete Sequenzlaenge ist
   unsigned int randlen();  // zufaellige Sequenzlaenge
   unsigned int explen();  // erwartete Sequenzlaenge
};


/*
class fbindel { // degenerierte Indelklasse fuer Pseudo-Hirschberg
 private:
   vector<vector<double> > vorwaerts;
   vector<vector<double> > rueckwaerts;
   vector<vector<double> > naht;
 public :
   fbindel(); // gibt Fehlermeldung
   fbindel(indel& id);
  ~fbindel();
   double lnforwprob(fbzustand f1, fbzustand f2);
   double lnbackwprob(fbzustand f1, fbzustand f2);
   double lncorrfac(fbzustand v, fbzustand r);
};
*/
#endif
