/****************************************************************************
 * This file is part of the program alifritz                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <cstdio>

using namespace std;

#include "seqcode.hh"
#include "parser.hh"


// functions which allow the parsing of the infile

scriptpart isnewpart(string Zeile)
{
   scriptpart retval=scNONE;
   string::size_type poscolon, poscis;
   
   poscolon=Zeile.find_first_of(':');
   poscis  =Zeile.find_first_of('#');
   
   if(poscolon!=string::npos && (poscis==string::npos || poscolon<poscis)) {  // neuer Abschnitt des Scripts
      string tillcolon=Zeile.substr(0,poscolon);
      if(findanycase(tillcolon,"SUBST"))
	    retval=SUBST;
      else {
         if(findanycase(tillcolon,"INDEL"))
	       retval=INDEL;
	 else {
            if(findanycase(tillcolon,"FILES"))
	          retval=FILES;
         }
      } 
   }
   
   return retval;
}



// falls im Teilscript SUBST
subst_option whichsubstoption(string Zeile, string& arg, option_arg& argtype)
{
   subst_option retval=suNONE;
   string::size_type poscis, posequ;
   string tillcis;   // Teilstring bis zum Kommentar
   string opt;       // Teilstring bis arg
   
   poscis = Zeile.find_first_of('#');
   if(poscis!=string::npos)
      tillcis = Zeile.substr(0,poscis);
   else
      tillcis = Zeile;
   posequ = tillcis.find_first_of('=');
   if(tillcis.size()>0 && posequ!=string::npos) { // gibt's ein '=' vor dem Kommentar
      arg=tillcis.substr(posequ+1,poscis-posequ-1);  // Argumentteil von tillcis
      argtype=typ_of_arg(arg);  // Was fuer ein Argument
      opt=tillcis.substr(0,posequ);    // Welche Option
      if(issubstmodel(opt))  // model=...
             retval = SMODEL;
      else {
             if(issubstparameter(opt))  // parameter=...
                    retval = SPARAM;
             else {
                    if(issubstfreqs(opt))  // freqs=...
                        retval = FREQS;
             }
      }
   }
   
   return retval;
}


bool issubstmodel(string opt)
{
   bool wf=findanycase(opt,"model");
   
   return wf;
}


bool issubstparameter(string opt)
{
   bool wf=findanycase(opt,"parameter");
   
   return wf;
}


bool issubstfreqs(string opt)
{
   bool wf=findanycase(opt,"freqs");
   
   return wf;
}



// falls im Teilscript INDEL
indel_option whichindeloption(string Zeile, string& arg, option_arg& argtype)
{
   indel_option retval=inNONE;
   string::size_type poscis, posequ;
   string tillcis;
   string opt;       // Teilstring bis arg
   
   poscis = Zeile.find_first_of('#');
   if(poscis!=string::npos)
      tillcis = Zeile.substr(0,poscis);
   else
      tillcis = Zeile;
   posequ = tillcis.find_first_of('=');
   if(tillcis.size()>0 && posequ!=string::npos) { // gibt's ein '=' vor dem Kommentar
      arg=tillcis.substr(posequ+1,poscis-posequ-1);  // Argumentteil von tillcis
      argtype=typ_of_arg(arg);  // Was fuer ein Argument
      opt=tillcis.substr(0,posequ);    // Welche Option
      if(isindmodel(opt))  // model=...
             retval = IMODEL;
      else {
	     if(isindparameter(opt))
	         retval = IPARAM;
      }
   }
   
   return retval;
}

bool isindmodel(string opt)
{
   bool wf=findanycase(opt,"model");
   
   return wf;
}


bool isindparameter(string opt)
{
   bool wf=findanycase(opt,"parameter");
   
   return wf;
}

//  falls im Teilscript FILES
files_option whichfilesoption(string Zeile, string& arg, option_arg& argtype)
{
   files_option retval=fiNONE;
   string::size_type poscis, posequ;
   string tillcis;
   string opt;       // Teilstring bis arg
   
   poscis = Zeile.find_first_of('#');
   if(poscis!=string::npos)
      tillcis = Zeile.substr(0,poscis);
   else
      tillcis = Zeile;
   posequ = tillcis.find_first_of('=');
   if(tillcis.size()>0 && posequ!=string::npos) { // gibt's ein '=' vor dem Kommentar
      arg=firstword(tillcis.substr(posequ+1,poscis-posequ-1));  // Argumentteil von tillcis
      argtype=typ_of_arg(arg);  // Was fuer ein Argument
      opt=tillcis.substr(0,posequ);    // Welche Option
      if(istreefile(opt))
             retval = TREEFILE;
      else {
	     if(isalifile(opt))
	            retval = ALIFILE;
	     else {
		    if(isseqfile(opt))
		           retval = SEQFILE;
	     }
      }
   }
   
   return retval;
}

bool istreefile(string opt)
{
   bool wf=findanycase(opt,"treefile");
   
   return wf;
}

bool isalifile(string opt)
{
   bool wf=findanycase(opt,"alifile");
   
   return wf;
}

bool isseqfile(string opt)
{
   bool wf=findanycase(opt,"seqfile");
   
   return wf;
}

// Wie sieht das Argument aus?
option_arg typ_of_arg(string& arg)
{
   option_arg retval=opNONE;
   string::size_type left_brack, right_brack, 
                     left_square, right_square;
   string newarg; 
   
   left_brack   = arg.find_first_of('(');
   right_brack  = arg.find_first_of(')');
   left_square  = arg.find_first_of('[');
   right_square = arg.find_first_of(']');
   
   if(left_brack!=string::npos) { // u.U. Vector
      if(right_brack==string::npos || right_brack<left_brack ||
         left_square!=string::npos || right_square!=string::npos) {
         cerr << "Invalid string " << arg << " in control file\n";
	 exit(EXIT_FAILURE);
      }
      retval=VECTOR;
      newarg=arg.substr(left_brack+1,right_brack-left_brack-1);
      arg=newarg;  // !!!!
   } else {
      if(left_square!=string::npos) { // u.U. [...]
         if(right_square==string::npos || right_square<left_square ||
            right_brack!=string::npos) {
            cerr << "Invalid string " << arg << " in control file\n";
	    exit(EXIT_FAILURE);
         }
         newarg=arg.substr(left_square+1,right_square-left_square-1);
         if(newarg.find("ALL")!=string::npos ||
	    newarg.find("all")!=string::npos ||
	    newarg.find("All")!=string::npos)
	         retval=ALL;
         else 
            retval=INTERVAL;
         
         arg=newarg;  // !!!!
      
      } else {
           if(arg.find("DEFAULT")!=string::npos ||
	      arg.find("default")!=string::npos ||
	      arg.find("Default")!=string::npos)
	         retval=DEFAULT;
           else {
              if(arg.find("ESTIMATE")!=string::npos ||
	         arg.find("estimate")!=string::npos ||
	         arg.find("Estimate")!=string::npos)
	            retval=ESTIMATE;
	      else 
	         retval=VALUE; // kann auch ein Dateiname sein
	   }
      }
   }
   
   return retval;
}

void stringtovector(string s, vector<double>& v)
{
   vector<double> fresh_vector;
   vector<string::size_type> commata;
   string::size_type compos;
   double d;

   compos = 0;
   if(s[0]==',') {
      cerr << "Invalid string " << s << " in function stringtovector\n";
      exit(EXIT_FAILURE);
   }
   do{ 
       compos=s.find(',',compos+1);
       if(compos!=string::npos)
          commata.push_back(compos);
       else
          commata.push_back(s.size());
   }while(compos!=string::npos);

   if(commata.size()==0) {
      if(sscanf(s.c_str(),"%lf",&d)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      fresh_vector.push_back(d);
   } else {
      // Anfang
      compos = 0;
      if(sscanf(s.substr(compos,commata[0]-compos).c_str(),"%lf",&d)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      compos=commata[0];
      fresh_vector.push_back(d);
      // Mittelteil und Ende
      for(unsigned int i = 1;i<commata.size();i++)
      {
         if(sscanf(s.substr(compos+1,commata[i]-compos-1).c_str(),"%lf",&d)!=1) 
         {
            cerr << "Cannot scan from string " << s << endl;
	    exit(EXIT_FAILURE);
         }
         compos=commata[i];
         fresh_vector.push_back(d);
      }
   }

   v=fresh_vector;
}

void stringtovector(string s, vector<unsigned long int>& v)
{
   vector<unsigned long int> fresh_vector;
   vector<string::size_type> commata;
   string::size_type compos;
   unsigned long int d;

   compos = 0;
   if(s[0]==',') {
      cerr << "Invalid string " << s << " in function stringtovector\n";
      exit(EXIT_FAILURE);
   }
   do{ 
       compos=s.find(',',compos+1);
       if(compos!=string::npos)
          commata.push_back(compos);
       else
          commata.push_back(s.size());
   }while(compos!=string::npos);

   if(commata.size()==0) {
      if(sscanf(s.c_str(),"%lu",&d)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      fresh_vector.push_back(d);
   } else {
      // Anfang
      compos = 0;
      if(sscanf(s.substr(compos,commata[0]-compos).c_str(),"%lu",&d)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      compos=commata[0];
      fresh_vector.push_back(d);
      // Mittelteil und Ende
      for(unsigned int i = 1;i<commata.size();i++)
      {
         if(sscanf(s.substr(compos+1,commata[i]-compos-1).c_str(),"%lu",&d)!=1) 
         {
            cerr << "Cannot scan from string " << s << endl;
	    exit(EXIT_FAILURE);
         }
         compos=commata[i];
         fresh_vector.push_back(d);
      }
   }

   v=fresh_vector;
}

void stringtolist(string s, list<int>& l)
{
   list<int> fresh_list;
   vector<string::size_type> commata;
   string::size_type compos;
   int ib;

   compos = 0;
   if(s[0]==',') {
      cerr << "Invalid string " << s << " in function stringtolist\n";
      exit(EXIT_FAILURE);
   }
   do{ 
       compos=s.find(',',compos+1);
       if(compos!=string::npos)
          commata.push_back(compos);
       else
          commata.push_back(s.size());
   }while(compos!=string::npos);

   if(commata.size()==0) {
      if(sscanf(s.c_str(),"%d",&ib)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      fresh_list.push_back(ib);
   } else {
      // Anfang
      compos = 0;
      if(sscanf(s.substr(compos,commata[0]-compos).c_str(),"%d",&ib)!=1) 
      {
         cerr << "Cannot scan from string " << s << endl;
	 exit(EXIT_FAILURE);
      }
      compos=commata[0];
      fresh_list.push_back(ib);
      // Mittelteil und Ende
      for(unsigned int i = 1;i<commata.size();i++)
      {
         if(sscanf(s.substr(compos+1,commata[i]-compos-1).c_str(),"%d",&ib)!=1) 
         {
            cerr << "Cannot scan from string " << s << endl;
	    exit(EXIT_FAILURE);
         }
         compos=commata[i];
         fresh_list.push_back(ib);
      }
   }

   l=fresh_list;
}



string oatostring(option_arg oa)
{
   switch(oa) {
      case opNONE :
        return "NONE";
      case DEFAULT :
        return "DEFAULT";
      case ESTIMATE :
        return "ESTIMATE";
      case VALUE :
        return "VALUE";
      case VECTOR :
        return "VECTOR";
      case INTERVAL :
        return "INTERVAL";
      case ALL :
        return "ALL";
      default :
        cerr << "Unknown value of oa\n";
	exit(EXIT_FAILURE);
   }
}


string firstword(string arg)
{
  bool flag=FALSE;
  string fw="";
  unsigned int i=0;
  
  while(i<arg.size()) {
     if(isprint(arg[i]) && !isspace(arg[i])) {
        flag=TRUE;
	fw+=arg[i];
     } else
       if(flag)
          break;
     i++;
  }

  return fw;
}



