/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <string>
#include <cmath>
#include <cfloat>
#include <cstdlib>
#include <iostream>

using namespace std;

#include "seqcode.hh"


void whichbase(vector<bool>& basbas, char c)
{
   unsigned int i;
   vector<int> zus;
   for(i=0;i<basbas.size();++i)
       basbas[i]=FALSE;
   whichwc(c,zus,(int)basbas.size());
   for(i=0;i<zus.size();++i)
       basbas[zus[i]]=TRUE;
}


int whichbase(char c,int alf)
{
   switch(alf) {
      case DNA :
         return chartobase(c);
      case PROTEIN :
         return chartoamac(c);
      default :
         cerr << "Unknown alphabet "<<alf<<" in function whichbase\n";
	 exit(EXIT_FAILURE);
   }
}


char state_to_char(int i,alphsize typ)
{
   switch(typ) {
      case DNA :
         return basetochar(i);
      case PROTEIN :
         return amactochar(i);
      default :
         cerr << "Unknown alphabet "<<(int)typ<<" in function state_to_char\n";
	 exit(EXIT_FAILURE);
   }
}


char basetochar(int i)
{
  char c;
  
  switch(i){
    case dAtp :
      c = 'A';
      break;
    case dCtp :
      c = 'C';
      break;
    case dGtp :
      c = 'G';
      break;
    case dTtp :
      c = 'T';
      break;
    case dRtp :
      c = 'R';
      break;
    case dYtp :
      c = 'Y';
      break;
    case dMtp :
      c = 'M';
      break;
    case dWtp :
      c = 'W';
      break;
    case dStp :
      c = 'S';
      break;
    case dKtp :
      c = 'K';
      break;
    case dBtp :
      c = 'B';
      break;
    case dDtp :
      c = 'D';
      break;
    case dHtp :
      c = 'H';
      break;
    case dVtp :
      c = 'V';
      break;
    case dNtp :
      c = 'N';
      break;
    default :
      cerr << "Unknown base " << i << " in function basetochar\n";
      exit(EXIT_FAILURE);
  }
  
  return c;
}

Base chartobase(char c)
{
   switch(c) {
         case 'A' :
            return dAtp;
	 case 'C' :
            return dCtp;
	 case 'G' :
            return dGtp;
	 case 'T' :
	 case 'U' :
            return dTtp;
	 case 'R' : // A || G
            return dRtp;
	 case 'Y' : // C || T
            return dYtp;
	 case 'M' : // A || C
            return dMtp;
	 case 'W' : // A || T
            return dWtp;
	 case 'S' : // C || G
            return dStp;
	 case 'K' : // G || T
            return dKtp;
	 case 'B' : // not A
            return dBtp;
	 case 'D' : // not C
            return dDtp;
	 case 'H' : // not G
            return dHtp;
	 case 'V' : // not T
            return dVtp;
	 case 'N' : // any nucleotide
            return dNtp;
	 default :
	   cerr << "Unknown symbol for base: " << c << " in function chartobase" << endl;
	   exit(EXIT_FAILURE);
  
     }
}

void chartobase(char c, vector<int>& bases)
{
   if(bases.size()!=0)
      bases.clear();
   
   switch(c) {
         case 'A' :
            bases.push_back(dAtp);
	 break;
	 case 'C' :
            bases.push_back(dCtp);
	 break;
	 case 'G' :
            bases.push_back(dGtp);
	 break;
	 case 'T' :
	 case 'U' :
            bases.push_back(dTtp);
	 break;
	 case 'R' : // A || G
            bases.push_back(dAtp);
	    bases.push_back(dGtp);
	 break;
	 case 'Y' : // C || T
            bases.push_back(dCtp);
	    bases.push_back(dTtp);
	 break;
	 case 'M' : // A || C
            bases.push_back(dAtp);
	    bases.push_back(dCtp);
	 break;
	 case 'W' : // A || T
            bases.push_back(dAtp);
	    bases.push_back(dTtp);
	 break;
	 case 'S' : // C || G
            bases.push_back(dCtp);
	    bases.push_back(dGtp);
	 break;
	 case 'K' : // G || T
            bases.push_back(dGtp);
	    bases.push_back(dTtp);
	 break;
	 case 'B' : // not A
            bases.push_back(dCtp);
            bases.push_back(dGtp);
	    bases.push_back(dTtp);
	 break;
	 case 'D' : // not C
            bases.push_back(dAtp);
            bases.push_back(dGtp);
	    bases.push_back(dTtp);
	 break;
	 case 'H' : // not G
            bases.push_back(dAtp);
            bases.push_back(dCtp);
	    bases.push_back(dTtp);
	 break;
	 case 'V' : // not T
            bases.push_back(dAtp);
            bases.push_back(dCtp);
	    bases.push_back(dGtp);
	 break;
	 case 'N' : // any nucleotide
            bases.push_back(dAtp);
            bases.push_back(dCtp);
	    bases.push_back(dGtp);
	    bases.push_back(dTtp);
	 break;
	 default :
            ;
     }
}

char amactochar(int i)
{
  char c;
  
  switch(i){
    case ALA :
       c = 'A';
       break;
    case CYS :
       c = 'C';
       break;
    case ASP :
       c = 'D';
       break;
    case GLU :
       c = 'E';
       break;
    case PHE :
       c = 'F';
       break;
    case GLY :
       c = 'G';
       break;
    case HIS :
       c = 'H';
       break;
    case ILE :
       c = 'I';
       break;
    case LYS :
       c = 'K';
       break;
    case LEU :
       c = 'L';
       break;
    case MET :
       c = 'M';
       break;
    case ASN :
       c = 'N';
       break;
    case PRO :
       c = 'P';
       break;
    case GLN :
       c = 'Q';
       break;
    case ARG :
       c = 'R';
       break;
    case SER :
       c = 'S';
       break;
    case THR :
       c = 'T';
       break;
    case VAL :
       c = 'V';
       break;
    case TRP :
       c = 'W';
       break;
    case TYR :
       c = 'Y';
       break;
    case ASX :
       c = 'B';
       break;
    case GLX :
       c = 'Z';
       break;
    case XXX :
       c = 'X';
       break;
    default :
      cerr << "Unknown amino acid " << i << " in function amactochar\n";
      exit(EXIT_FAILURE);
  }
  
  return c;
}


void chartoamac(char c, vector<int>& amacs)
{
   if(amacs.size()!=0)
      amacs.clear();
   
   switch(c) {
         case 'A' :
            amacs.push_back(ALA);
	 break;
         case 'B' :
            amacs.push_back(ASN);
            amacs.push_back(ASP);
	 break;
	 case 'C' :
            amacs.push_back(CYS);
	 break;
	 case 'D' :
            amacs.push_back(ASP);
	 break;
	 case 'E' :
            amacs.push_back(GLU);
	 break;
	 case 'F' :
            amacs.push_back(PHE);
	 break;
	 case 'G' :
            amacs.push_back(GLY);
	 break;
	 case 'H' :
            amacs.push_back(HIS);
	 break;
	 case 'I' :
            amacs.push_back(ILE);
	 break;
	 case 'K' :
            amacs.push_back(LYS);
	 break;
	 case 'L' :
            amacs.push_back(LEU);
	 break;
	 case 'M' :
            amacs.push_back(MET);
	 break;
	 case 'N' :
            amacs.push_back(ASN);
	 break;
	 case 'P' :
            amacs.push_back(PRO);
	 break;
	 case 'Q' :
            amacs.push_back(GLN);
	 break;
	 case 'R' :
            amacs.push_back(ARG);
	 break;
	 case 'S' :
            amacs.push_back(SER);
	 break;
	 case 'T' :
            amacs.push_back(THR);
	 break;
	 case 'V' :
            amacs.push_back(VAL);
	 break;
	 case 'W' :
            amacs.push_back(TRP);
	 break;
	 case 'X' :
            amacs.push_back(ALA);
	    amacs.push_back(ARG);
	    amacs.push_back(ASN);
            amacs.push_back(ASP);
	    amacs.push_back(CYS);
	    amacs.push_back(GLN);
            amacs.push_back(GLU);
	    amacs.push_back(GLY);
	    amacs.push_back(HIS);
	    amacs.push_back(ILE);
	    amacs.push_back(LEU);
	    amacs.push_back(LYS);
	    amacs.push_back(MET);
	    amacs.push_back(PHE);
	    amacs.push_back(PRO);
	    amacs.push_back(SER);
	    amacs.push_back(THR);
	    amacs.push_back(TRP);
            amacs.push_back(TYR);
	    amacs.push_back(VAL);
	 break;
	 case 'Y' :
            amacs.push_back(TYR);
	 break;
	 case 'Z' :
            amacs.push_back(GLN);
            amacs.push_back(GLU);
	 break;
	 default :
            ;
   }
}

AminoAcid chartoamac(char c)
{
   switch(c) {
         case 'A' :
            return ALA;
	 case 'B' :
            return ASX;
	 case 'C' :
            return CYS;
	 case 'D' :
            return ASP;
	 case 'E' :
            return GLU;
	 case 'F' :
            return PHE;
	 case 'G' :
            return GLY;
	 case 'H' :
            return HIS;
	 case 'I' :
            return ILE;
	 case 'K' :
            return LYS;
	 case 'L' :
            return LEU;
	 case 'M' :
            return MET;
	 case 'N' :
            return ASN;
	 case 'P' :
            return PRO;
	 case 'Q' :
            return GLN;
	 case 'R' :
            return ARG;
	 case 'S' :
            return SER;
	 case 'T' :
            return THR;
	 case 'V' :
            return VAL;
	 case 'W' :
            return TRP;
	 case 'X' :
            return XXX;
	 case 'Y' :
            return TYR;
	 case 'Z' :
            return GLX;
	 default :
	   cerr << "Unknown symbol for amino acid: " << c << " in function chartoamac" << endl;
	   exit(EXIT_FAILURE);
   }
}


void whichwildcard(int wc,vector<bool>& zus)
{
   unsigned int alf=zus.size();
   unsigned int i;
   for(i=0;i<alf;++i)
       zus[i]=FALSE;
   switch(alf) {
      case DNA :
   	   switch(wc) {
              case dAtp :
	      case dCtp :
	      case dGtp :
	      case dTtp :
	         zus[wc]=TRUE;
                 break;
	      case dRtp : // A || G
	         zus[dAtp]=zus[dGtp]=TRUE;
                 break;
	      case dYtp : // C || T
	         zus[dCtp]=zus[dTtp]=TRUE;
                 break;
	      case dMtp : // A || C
	         zus[dAtp]=zus[dCtp]=TRUE;
                 break;
	      case dWtp : // A || T
	         zus[dAtp]=zus[dTtp]=TRUE;
                 break;
	      case dStp : // C || G
	         zus[dCtp]=zus[dGtp]=TRUE;
                 break;
	      case dKtp : // G || T
	         zus[dGtp]=zus[dTtp]=TRUE;
                 break;
	      case dBtp : // not A
	         zus[dCtp]=zus[dGtp]=zus[dTtp]=TRUE;
                 break;
	      case dDtp : // not C
	         zus[dAtp]=zus[dGtp]=zus[dTtp]=TRUE;
                 break;
	      case dHtp : // not G
	         zus[dAtp]=zus[dCtp]=zus[dTtp]=TRUE;
                 break;
	      case dVtp : // not T
	         zus[dAtp]=zus[dCtp]=zus[dGtp]=TRUE;
                 break;
	      case dNtp : // any nucleotide
	         for(i=0;i<alf;++i)
		    zus[i]=TRUE;
                 break;
	      default :
	        cerr << "Unknown base " << wc << " in function whichwildcard" << endl;
	        exit(EXIT_FAILURE);
           }
           break;
      case PROTEIN :
           switch(wc) {
              case ALA :
	      case CYS :
	      case ASP :
	      case GLU :
	      case PHE :
	      case GLY :
	      case HIS :
	      case ILE :
	      case LYS :
	      case LEU :
	      case MET :
	      case ASN :
	      case PRO :
	      case GLN :
	      case ARG :
	      case SER :
	      case THR :
	      case VAL :
	      case TRP :
	      case TYR :
	         zus[wc]=TRUE;
                 break;
	      case ASX :
	         zus[ASP]=zus[ASN]=TRUE;
                 break;
	      case GLX :
	         zus[GLU]=zus[GLN]=TRUE;
                 break;
	      case XXX :
	         for(i=0;i<alf;++i)
		    zus[i]=TRUE;
                 break;
	      default :
	        cerr << "Unknown amino acid: " << wc << " in function whichwildcard" << endl;
	        exit(EXIT_FAILURE);
           }
           break;
      default :
        cerr << "Unknown alphabet in function whichwildcard\n";
	exit(EXIT_FAILURE);
   }
}

void whichwc(char wc,vector<int>& zus, int alf)
{
   zus.clear();
   switch(alf) {
      case DNA :
   	   switch(wc) {
              case 'A' :
	         zus.push_back(dAtp);
                 break;
	      case 'C' :
	         zus.push_back(dCtp);
                 break;
	      case 'G' :
	         zus.push_back(dGtp);
                 break;
	      case 'T' :
	         zus.push_back(dTtp);
                 break;
	      case 'R' : // A || G
	         zus.push_back(dAtp);
	         zus.push_back(dGtp);
                 break;
	      case 'Y' : // C || T
	         zus.push_back(dCtp);
	         zus.push_back(dTtp);
                 break;
	      case 'M' : // A || C
	         zus.push_back(dAtp);
	         zus.push_back(dCtp);
                 break;
	      case 'W' : // A || T
	         zus.push_back(dAtp);
	         zus.push_back(dTtp);
                 break;
	      case 'S' : // C || G
	         zus.push_back(dCtp);
	         zus.push_back(dGtp);
                 break;
	      case 'K' : // G || T
	         zus.push_back(dGtp);
	         zus.push_back(dTtp);
                 break;
	      case 'B' : // not A
	         zus.push_back(dCtp);
	         zus.push_back(dGtp);
	         zus.push_back(dTtp);
                 break;
	      case 'D' : // not C
	         zus.push_back(dAtp);
	         zus.push_back(dGtp);
	         zus.push_back(dTtp);
                 break;
	      case 'H' : // not G
	         zus.push_back(dAtp);
	         zus.push_back(dCtp);
	         zus.push_back(dTtp);
                 break;
	      case 'V' : // not T
	         zus.push_back(dAtp);
	         zus.push_back(dCtp);
	         zus.push_back(dGtp);
                 break;
	      case 'N' : // any nucleotide
	         for(int i=0;i<alf;++i)
		    zus.push_back(i);
                 break;
	      default :
	        cerr << "Unknown base " << wc << " in function whichwc" << endl;
	        exit(EXIT_FAILURE);
           }
           break;
      case PROTEIN :
           switch(wc) {
              case 'A' :
	         zus.push_back(ALA);
                 break;
	      case 'C' :
	         zus.push_back(CYS);
                 break;
	      case 'D' :
	         zus.push_back(ASP);
                 break;
	      case 'E' :
	         zus.push_back(GLU);
                 break;
	      case 'F' :
	         zus.push_back(PHE);
                 break;
	      case 'G' :
	         zus.push_back(GLY);
                 break;
	      case 'H' :
	         zus.push_back(HIS);
                 break;
	      case 'I' :
	         zus.push_back(ILE);
                 break;
	      case 'K' :
	         zus.push_back(LYS);
                 break;
	      case 'L' :
	         zus.push_back(LEU);
                 break;
	      case 'M' :
	         zus.push_back(MET);
                 break;
	      case 'N' :
	         zus.push_back(ASN);
                 break;
	      case 'P' :
	         zus.push_back(PRO);
                 break;
	      case 'Q' :
	         zus.push_back(GLN);
                 break;
	      case 'R' :
	         zus.push_back(ARG);
                 break;
	      case 'S' :
	         zus.push_back(SER);
                 break;
	      case 'T' :
	         zus.push_back(THR);
                 break;
	      case 'V' :
	         zus.push_back(VAL);
                 break;
	      case 'W' :
	         zus.push_back(TRP);
                 break;
	      case 'Y' :
	         zus.push_back(TYR);
                 break;
	      case 'B' :
	         zus.push_back(ASP);
	         zus.push_back(ASN);
                 break;
	      case 'Z' :
	         zus.push_back(GLU);
	         zus.push_back(GLN);
                 break;
	      case 'X' :
	         for(int i=0;i<alf;++i)
		    zus.push_back(i);
                 break;
	      default :
	        cerr << "Unknown amino acid: " << wc << " in function whichwc" << endl;
	        exit(EXIT_FAILURE);
           }
           break;
      default :
        cerr << "Unknown alphabet in function whichwc\n";
	exit(EXIT_FAILURE);
   }
}

/*
char profile_to_char(vector<double>& prof)
{
   switch(prof.size()) {
      case 4 :
         return basprof_tochar(prof);
      case 20 :
         return amacprof_tochar(prof);
      default :
         cerr << "Unknown profile type with "<<prof.size()<< " states\n";
	 exit(EXIT_FAILURE);
   }
}

char basprof_tochar(vector<double>& prof)
{
    vector<bool> bc(prof.size());
    for(unsigned int i=0;i<prof.size();++i){
        if(prof[i]>0.0+DBL_MIN)
	   bc[i]=TRUE;
	else
	   bc[i]=FALSE;    
    }
    switch(bc[dAtp]) {
       case TRUE :
          switch(bc[dCtp]) {
	     case TRUE :
	       switch(bc[dGtp]) {
	          case TRUE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'N';
		       case FALSE :
		         return 'V';
		       default :
		         cerr << "Impossible I";
			 exit(EXIT_FAILURE);
		    }
		  case FALSE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'H';
		       case FALSE :
		         return 'M';
		       default :
		         cerr << "Impossible II";
			 exit(EXIT_FAILURE);
		    }
		  default :
		    cerr << "Impossible III";
		    exit(EXIT_FAILURE);
	       }
	     case FALSE :
	       switch(bc[dGtp]) {
	          case TRUE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'D';
		       case FALSE :
		         return 'R';
		       default :
		         cerr << "Impossible IV";
			 exit(EXIT_FAILURE);
		    }
		  case FALSE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'W';
		       case FALSE :
		         return 'A';
		       default :
		         cerr << "Impossible V";
			 exit(EXIT_FAILURE);
		    }
		  default :
		    cerr << "Impossible VI";
		    exit(EXIT_FAILURE);
	       }
	     default :
	       cerr << "Impossible VII";
	       exit(EXIT_FAILURE);
	  }
       case FALSE :
          switch(bc[dCtp]) {
	     case TRUE :
	       switch(bc[dGtp]) {
	          case TRUE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'B';
		       case FALSE :
		         return 'S';
		       default :
		         cerr << "Impossible VIII";
			 exit(EXIT_FAILURE);
		    }
		  case FALSE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'Y';
		       case FALSE :
		         return 'C';
		       default :
		         cerr << "Impossible IX";
			 exit(EXIT_FAILURE);
		    }
		  default :
		    cerr << "Impossible X";
		    exit(EXIT_FAILURE);
	       }
	     case FALSE :
	       switch(bc[dGtp]) {
	          case TRUE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'K';
		       case FALSE :
		         return 'G';
		       default :
		         cerr << "Impossible XI";
			 exit(EXIT_FAILURE);
		    }
		  case FALSE :
		    switch(bc[dTtp]) {
		       case TRUE :
		         return 'T';
		       case FALSE :
		         cerr << "Could not recover base from profile\n";
			 exit(EXIT_FAILURE);
		       default :
		         cerr << "Impossible XII";
			 exit(EXIT_FAILURE);
		    }
		  default :
		    cerr << "Impossible XIII";
		    exit(EXIT_FAILURE);
	       }
	     default :
	       cerr << "Impossible XIV";
	       exit(EXIT_FAILURE);
	  }
       default :
          cerr << "Impossible XV";
	  exit(EXIT_FAILURE);
    }
}

char amacprof_tochar(vector<double>& prof)
{
    vector<int> amacidx;
    
    for(unsigned int i=0;i<prof.size();++i)
       if(prof[i]>0.0+DBL_MIN)
          amacidx.push_back(i);
	  
     switch(amacidx.size()) {
        case 20 :
	   return 'X';
	case 2 :
	   if((amacidx[0]==ASN && amacidx[1]==ASP)||(amacidx[0]==ASP && amacidx[1]==ASN))
	     return 'B';
	   if((amacidx[0]==GLN && amacidx[1]==GLU)||(amacidx[0]==GLU && amacidx[1]==GLN))
	     return 'Z';
	   cerr << "The profile is so strange that I'd better exit\n";
	   exit(EXIT_FAILURE);
	case 1 :
	   return amactochar(amacidx[0]);
	default :
	   cerr << "The profile is so strange that I'd better exit\n";
	   exit(EXIT_FAILURE);
     }
}
*/

int alftologalf(int alf) // Wieviele Zustaende (auch unsichere) gibt's?
{
   switch(alf) {
       case DNA :
          return 15;
       case PROTEIN :
          return 23;
       default :
          cerr << "Unknown alphabet "<<(int)alf<<endl;
	  exit(EXIT_FAILURE);
   }
}

subarch stringtosubarch(string arg)
{
   if(findanycase(arg,"GENREV"))
      return GENREV;
   if(findanycase(arg,"TAMNEI"))
      return TAMNEI;
   if(findanycase(arg,"HKY"))
      return HKY;
   if(findanycase(arg,"KIMURA2P"))
      return KIMURA2P;
   if(findanycase(arg,"KIMURA3P"))
      return KIMURA3P;
   if(findanycase(arg,"FEL81"))
      return FEL81;
   if(findanycase(arg,"JC69"))
      return JC69;
   if(findanycase(arg,"MV2000"))
      return MV2000;
   if(findanycase(arg,"JTT"))
      return JTT;
   if(findanycase(arg,"DAYHOFF"))
      return DAYHOFF;
   if(findanycase(arg,"MTREV24"))
      return MTREV24;
   if(findanycase(arg,"BLOSUM62"))
      return BLOSUM62;
   if(findanycase(arg,"SH94"))
      return SH94;
   
   cerr << "No key word was found where the name of\n"
        << "the substitution model was expected.\n";
   exit(EXIT_FAILURE);
}


idarch stringtoidarch(string arg)
{
   if(findanycase(arg,"TKF1"))
      return TKF1;
   if(findanycase(arg,"TKF2"))
      return TKF2;
   if(findanycase(arg,"FID"))
      return FID;

   cerr << "No key word was found where the name of the\n"
        << "insertion-deletion model was expected.\n";
   exit(EXIT_FAILURE);
}


bool findanycase(string text, string pattern)
{
   unsigned int i;
   
   string actext="";
   string acpattern="";
   for(i=0;i<text.size();i++)
       actext+=toupper(text[i]);
   for(i=0;i<pattern.size();i++)
       acpattern+=toupper(pattern[i]);

   if(actext.find(acpattern)!=string::npos)
       return TRUE;
   else
       return FALSE;
}


