/****************************************************************************
 * This file is part of the program DOAL                                    *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <iostream>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <ctime>
#include <string>
#include <list>
#include <climits>
#include <cfloat>

using namespace std;

#include "seqcode.hh"
#include "parser.hh"
#include "eingabe.hh"
#include "indel.hh"
#include "substmod.hh"
#include "treespec.hh"
#include "data.hh"
#include "random.hh"

#define MINDESTLAENGE 10

int main(int argc,char* argv[]) {

    initrandom();
    
    // Eingabeparameter einlesen
    if(argc!=4){
       cerr << "Please invoke simulator with the name of the control file,\n"
            <<"the sequence length and the number of simulations as arguments.\n";
       exit(EXIT_FAILURE);
    }
    eingabe_parameter ep;
    ep.from_file(argv[1]);

    substmodel sub(ep.sub.retalf(),ep.sub.model,ep.sub.retfreqs(),ep.sub.params);
    indel id(ep.ind.params,ep.ind.model);
    unsigned int explen, numdat;
    if(sscanf(argv[2],"%u",&explen)!=1) {
       cerr << "Could not read the expected sequence length\n";
       exit(EXIT_FAILURE);
    }
    if(explen<MINDESTLAENGE){
        cerr << "I don't want to simulate sequences which are that short!\n";
	exit(EXIT_FAILURE);
    }
    if(sscanf(argv[3],"%u",&numdat)!=1) {
       cerr << "Could not read the number of datasets\n";
       exit(EXIT_FAILURE);
    }
    id.setexplength(explen);  // passt mue so an, dass die gewuenschte Sequenzlaenge die erwartete ist
    clustal bstr(ep.fil.treefile.c_str());
    baum *tree = new baum(bstr,TRUE);

    ofstream outali(ep.fil.alifile.c_str());
    if(!outali) {
      cerr << "Cannot open output file "<<ep.fil.alifile.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    ofstream outseqs(ep.fil.seqfile.c_str());
    if(!outseqs) {
      cerr << "Cannot open output file "<<ep.fil.seqfile.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    
    for(unsigned int i=0;i<numdat;++i) {
       int Maxseqc=0;
       bool flag = FALSE;
       do{
          flag=tree->evolve(sub,id);
	  if(flag) {  // an allen Blaettern muessen auch Sequenzen vorhanden sein
             dataset data(tree,sub.retalf());
	     Maxseqc=data.laen(0);
	     if(Maxseqc>=MINDESTLAENGE) {
                data.print_phylip(outali);
                data.print_fasta(outseqs);
	     }
	  }
          tree->clear();
       } while(Maxseqc<MINDESTLAENGE); // das Alignment darf nicht kuerzer als MINDESTLAENGE sein
    }
    
    delete tree;
    outali.close();
    outseqs.close();
    
    return 0;
}

