/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <string>
#include <iostream>
#include <cfloat>
#include <cmath>
#include <cstdlib>

using namespace std;

#include "seqcode.hh"
#include "substmod.hh"
#include "util.h"
#include "random.hh"

#define INFACC 100.0

double untidy2_addlog(double a, double b) {  // zweimal im Programm vorhanden!!! (aufraeumen)
    // liefert log(exp(a)+exp(b))
    if (a>b) return a+log(1+exp(b-a));
    else return b+log(1+exp(a-b));
}


double untidy2_addlog(vector<double>& v) {
    // liefert log(Summe_ueber_i von exp(v[i]))
    if(v.size()==0) {
       cerr << "Not enough arguments to addlog\n";
       exit(EXIT_FAILURE);
    }
    if(v.size()==1)
       return v[0];
    if(v.size()==2)
       return untidy2_addlog(v[0],v[1]);
    vector<double> v2 = v;
    v2.pop_back();
    return untidy2_addlog(untidy2_addlog(v2),v.back());
}


/**********************************************************************
 *******************    class substmodel     **************************
 **********************************************************************/

substmodel::substmodel()
{
  int i;
  
  alf=DNA;
  logalf=alftologalf(alf);   
  Eval.resize(alf);
  Evec.resize(alf);
  Ievc.resize(alf);
  iexp.resize(alf);
  statefreqs.resize(alf);
  //logstatefreqs.resize(logalf);
  rates.resize(alf);
  probs_time.resize(alf);
  //logprobs_time.resize(logalf);
  for(i=0;i<alf;++i) {
     rates[i].resize(alf);
     probs_time[i].resize(alf);
     //logprobs_time[i].resize(logalf);
     Evec[i].resize(alf);
     Ievc[i].resize(alf);
     iexp[i].resize(alf);
  }
  //for(i=alf;i<logalf;++i)
     //logprobs_time[i].resize(logalf);
  
  model = JC69;

  time=1.0;
  tranprobmat();
  tprobmtrx(); 
}

substmodel::substmodel(alphsize a)
{ // default: Jukes-Cantor or Dayhoff
  int i;
  alf=a;
  logalf=alftologalf(alf);   
  Eval.resize(alf);
  Evec.resize(alf);
  Ievc.resize(alf);
  iexp.resize(alf);
  statefreqs.resize(alf);
  //logstatefreqs.resize(logalf);
  rates.resize(alf);
  probs_time.resize(alf);
  //logprobs_time.resize(logalf);
  for(i=0;i<alf;++i) {
     rates[i].resize(alf);
     probs_time[i].resize(alf);
     //logprobs_time[i].resize(logalf);
     Evec[i].resize(alf);
     Ievc[i].resize(alf);
     iexp[i].resize(alf);
  }
  //for(i=alf;i<logalf;++i)
     //logprobs_time[i].resize(logalf);
  
  switch(alf) {
    case DNA :
       model = JC69;
       break;
    case PROTEIN :
       model = DAYHOFF;
       break;
    default :
       cerr << "The program can only deal with nucleotide and\n"
            << "amino acid sequences.\n";
       exit(EXIT_FAILURE);
  }

  time=1.0;
  tranprobmat();
  tprobmtrx(); 
}

substmodel::substmodel(alphsize a, subarch m)
{
  int i;
  alf=a;
  logalf=alftologalf(alf);   

  Eval.resize(alf);
  Evec.resize(alf);
  Ievc.resize(alf);
  iexp.resize(alf);
  statefreqs.resize(alf);
  //logstatefreqs.resize(logalf);
  rates.resize(alf);
  probs_time.resize(alf);
  //logprobs_time.resize(logalf);
  for(i=0;i<alf;++i) {
     rates[i].resize(alf);
     probs_time[i].resize(alf);
     //logprobs_time[i].resize(logalf);
     Evec[i].resize(alf);
     Ievc[i].resize(alf);
     iexp[i].resize(alf);
  }
  //for(i=alf;i<logalf;++i)
     //logprobs_time[i].resize(logalf);
  
  model=m;
  
  switch(model) {
    case FEL81 :
       cout << "The Felsenstein model needs base frequencies.\n"
            << "The Jukes Cantor model will be used instead.\n";
       model=JC69;
    case JC69 :
       break;
    case HKY :
       cout << "The HKY model needs base frequencies.\n"
            << "Kimura2P will be used instead.\n";
       model=KIMURA2P;
    case KIMURA2P :
       params.resize(1);
       params[0]=0.5;
       break;
    case GENREV :
    case TAMNEI :
       cout << "The Tamura Nei model and the general reversible model\n"
            << "need base frequencies. Kimura3P will be used instead.\n";
       model=KIMURA3P;
    case KIMURA3P :
       params.resize(2);
       params[0]=0.5;
       params[1]=1.0;
       break;
    case MV2000 :
       cout << "The general reversible model for proteins needs amino acid frequencies.\n"
            << "The Dayhoff model will be used instead.\n";
       model=DAYHOFF;
    case DAYHOFF :
    case BLOSUM62 :
    case JTT :
    case MTREV24 :
       break;   
    default :
       cerr << "Unknown substitution model\n";
       exit(EXIT_FAILURE);
  }

  time=1.0;
  tranprobmat();
  tprobmtrx(); 
}

substmodel::substmodel(alphsize a, subarch m, vector<double>& freqs)
{
  alf=a;
  logalf=alftologalf(alf);   
  int i;
  double fsum;

  statefreqs.resize(alf);
  if(statefreqs.size()!=freqs.size()){
    cerr << "The data type has " << statefreqs.size() << " states,\n"
         << "yet there are only " << freqs.size() << " frequencies.\n";
    exit(EXIT_FAILURE);
  }
  fsum = 0.0;
  for(i=0;i<alf;++i) {
     if(freqs[i]<0.0+DBL_MIN) {
        cerr << "The frequency of at least one state is not positive.\n";
        exit(EXIT_FAILURE);
     }
     fsum+=freqs[i];
  }
  
  //logstatefreqs.resize(logalf);
  Eval.resize(alf);
  Evec.resize(alf);
  Ievc.resize(alf);
  iexp.resize(alf);
  rates.resize(alf);
  probs_time.resize(alf);
  //logprobs_time.resize(logalf);
  for(i=0;i<alf;++i) {
     rates[i].resize(alf);
     probs_time[i].resize(alf);
     //logprobs_time[i].resize(logalf);
     Evec[i].resize(alf);
     Ievc[i].resize(alf);
     iexp[i].resize(alf);
  }
  //for(i=alf;i<logalf;++i)
     //logprobs_time[i].resize(logalf);
  
  model=m;
  
  switch(model) {
    case FEL81 :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case JC69 :
       break;
    case HKY :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case KIMURA2P :
       params.resize(1);
       params[0]=0.5;
       break;
    case GENREV :
       params.resize(6);
       for(i=0;i<(int)params.size();++i)
         params[i]=1.0;
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
       break;
    case TAMNEI :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case KIMURA3P :
       params.resize(2);
       params[0]=0.5;
       params[1]=1.0;
       break;
    case MV2000 :
       params.resize(190);
       for(i=0;i<(int)params.size();++i)
         params[i]=1.0;
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case DAYHOFF :
    case BLOSUM62 :
    case JTT :
    case MTREV24 :
       break;   
    default :
       cerr << "Unknown substitution model\n";
       exit(EXIT_FAILURE);
  }

  time=1.0;
  tranprobmat();
  tprobmtrx(); 
}

substmodel::substmodel(alphsize a, subarch m, vector<double>& freqs, vector<double>& parameter)
{
  alf=a;
  logalf=alftologalf(alf);   
  int i;
  double fsum;

  statefreqs.resize(alf);
  if(statefreqs.size()!=freqs.size()){
    cerr << "The data type has " << statefreqs.size() << " states,\n"
         << "yet there are only " << freqs.size() << " frequencies.\n";
    exit(EXIT_FAILURE);
  }
  fsum = 0.0;
  for(i=0;i<alf;++i) {
     if(freqs[i]<0.0+DBL_MIN) {
        cerr << "The frequency of at least one state is not positive.\n";
        exit(EXIT_FAILURE);
     }
     fsum+=freqs[i];
  }
  //logstatefreqs.resize(logalf);
  Eval.resize(alf);
  Evec.resize(alf);
  Ievc.resize(alf);
  iexp.resize(alf);
  rates.resize(alf);
  probs_time.resize(alf);
  //logprobs_time.resize(logalf);
  for(i=0;i<alf;++i) {
     rates[i].resize(alf);
     probs_time[i].resize(alf);
     //logprobs_time[i].resize(logalf);
     Evec[i].resize(alf);
     Ievc[i].resize(alf);
     iexp[i].resize(alf);
  }
  //for(i=alf;i<logalf;++i)
     //logprobs_time[i].resize(logalf);
  
  model=m;
  
  switch(model) {
    case FEL81 :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case JC69 :
       if(parameter.size()>0)
         cerr << "The Felsenstein model and the Jukes Cantor do not use\n"
	      << "any rate parameters. They will be ignored!\n";
       break;
    case HKY :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case KIMURA2P :
       params.resize(1);
       if(params.size()!=parameter.size()) {
         cerr << "The HKY and the K2P model use only one rate parameter.\n"
	      << "The program will use the default value 0.5.\n"; 
         params[0]=0.5;
       } else 
         params[0]=parameter[0];
       break;
    case GENREV :
       params.resize(6);
       if(params.size()!=parameter.size()) {
         cerr << "The general reversible model uses six rate parameters.\n"
	      << "By default each rate is set to 1.0.\n";
         for(i=0;i<(int)params.size();++i)
           params[i]=1.0;
       } else
         for(i=0;i<(int)params.size();++i)
           params[i]=parameter[i];
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
       break;
    case TAMNEI :
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case KIMURA3P :
       params.resize(2);
       if(params.size()!=parameter.size()) {
         cerr << "The Tamura Nei and the K3P model use two rate parameters.\n"
	      << "By default they are set to 0.5 and 1.0.\n";
         params[0]=0.5;
         params[1]=1.0;
       } else
         for(i=0;i<(int)params.size();++i)
           params[i]=parameter[i];
       break;
    case MV2000 :
       params.resize(190);
       if(params.size()!=parameter.size()) {
         cerr << "The general reversible model for amino acids uses 190 rate parameters.\n"
	      << "By default each rate is set to 1.0.\n";
         for(i=0;i<(int)params.size();++i)
           params[i]=1.0;
       } else
         for(i=0;i<(int)params.size();++i)
           params[i]=parameter[i];
       for(i=0;i<alf;++i)
         statefreqs[i]=freqs[i]/fsum;
    case DAYHOFF :
    case BLOSUM62 :
    case JTT :
    case MTREV24 :
       break;   
    default :
       cerr << "Unknown substitution model\n";
       exit(EXIT_FAILURE);
  }

  time=1.0;
  tranprobmat();
  tprobmtrx(); 
}

substmodel::substmodel(const substmodel& s)
{
  params=s.params;
  model=s.model;
  alf=s.alf;   
  logalf=s.logalf;   
  statefreqs=s.statefreqs;
  //logstatefreqs=s.logstatefreqs;
  rates=s.rates;
  time=s.time;
  probs_time=s.probs_time;
  //logprobs_time=s.logprobs_time;
  Eval=s.Eval;
  Evec=s.Evec;
  Ievc=s.Ievc;
  iexp=s.iexp;
}

substmodel::~substmodel()
{
}


void substmodel::settime(double ptime)
{
  time = ptime;
  tprobmtrx();
}

void substmodel::setfreqs(vector<double>& freqs)
{
  unsigned int i;
  double fsum;
  
  if(statefreqs.size()!=freqs.size()){
    cerr << "The data type has " << statefreqs.size() << " states,\n"
         << "yet there are only " << freqs.size() << " frequencies.\n";
    exit(EXIT_FAILURE);
  }
  fsum = 0.0;
  for(i=0;i<freqs.size();++i) {
     if(freqs[i]<0.0+DBL_MIN) {
        cerr << "The frequency of at least one state is not positive.\n";
        exit(EXIT_FAILURE);
     }
     fsum+=freqs[i];
  }
  
  switch(model) {
    case GENREV :
    case TAMNEI :
    case HKY :
    case FEL81 :
    case MV2000 :
       break;
    case KIMURA2P :
    case KIMURA3P :
    case JC69 :
    case JTT :
    case DAYHOFF :
    case MTREV24 :
    case BLOSUM62 :
       cerr << "You were trying to change the base or amino acid frequencies\n";
       exit(EXIT_FAILURE);
       break;
    default :
       cerr << "Unknown substitution model\n";
       exit(EXIT_FAILURE);
  }
  for(i=0;i<freqs.size();++i)
     statefreqs[i]=freqs[i];
  

  tranprobmat();
  tprobmtrx();
}

void substmodel::setparams(vector<double>& p)
{
  unsigned int i;

  if(p.size()!=params.size()) {
     cerr << "Inadequate number of parameters in function setparams\n";
     exit(EXIT_FAILURE);
  }
  for(i=0;i<p.size();++i)
     params[i]=p[i];

  tranprobmat();
  tprobmtrx();
}

double substmodel::timeOf()
{
   return time;
}

void substmodel::freqsOf(vector<double>& f)
{
  int i;
  
  f.resize(alf);
  for(i=0;i<alf;++i)
    f[i]=statefreqs[i];
}

subarch substmodel::paramsOf(vector<double>& p)
{
  unsigned int i;
  
  p.resize(params.size());
  for(i=0;i<p.size();++i)
    p[i]=params[i];
    
  return model;
}

double substmodel::logfreqofstate(int state)
{
   if(state<0 || state>=logalf) {
      cerr << "Indexing error in logfreqofstate: idx="<<state<<endl;
      exit(EXIT_FAILURE);
   }
   if(state<alf)
      return log(statefreqs[state]);
   double probsum=0.0;
   vector<bool> zus(alf);
   whichwildcard(state,zus);
   for(int j=0;j<alf;++j)
       if(zus[j])
          probsum+=statefreqs[j];
   return log(probsum);
}
/*double substmodel::freqofstate(char state)
{
   double freq;
   
   switch (alf) {
      case DNA :
        freq=basefreq(state);
	break;
      case PROTEIN :
        freq=amacfreq(state);
	break;
      default :
        cerr << "Unknown data type in function freqofstate\n";
	exit(EXIT_FAILURE);
   }
   
   return freq;
}

double substmodel::Erw_freqofstate(const vector<double>&prof)
{ // gewichtet mit Profil
   double freq;
   
   if((int)prof.size()!=alf) {
      cerr << "The profile does not have the specified data type!\n";
      exit(EXIT_FAILURE);
   }

  freq = 0.0;  
  for(int i=0;i<alf;++i)
     freq += prof[i]*statefreqs[i];
    
   return freq;
}

double substmodel::basefreq(char base)
{
  double freq;
  vector<int> b;
  
  if(base=='N')
    freq=1.0;
  else {
    chartobase(base,b);
  
    freq = 0.0;
    for(unsigned int i=0;i<b.size();++i)
       freq += statefreqs[b[i]];
  }
    
  return freq;
}

double substmodel::amacfreq(char amac)
{
  double freq;
  vector<int> a;
  
  if(amac=='X')
    freq=1.0;
  else {
    chartoamac(amac,a);

    freq = 0.0;
    for(unsigned int i=0;i<a.size();++i)
       freq += statefreqs[a[i]];
  }
  
  return freq;
}
*/
double substmodel::logProb_ij(int i, int j)
{
  /* gibt den i,j-ten Eintrag in der logW'keitsmatrix
     zurueck */
  if(i>=logalf || j>=logalf)
  {
    cerr << "Indices too big: i = " << i << ", j = " << j << "\n";
    exit(EXIT_FAILURE);
  }
  if(i<alf && j<alf)
     return log(probs_time[i][j]);
  if(i<alf) {
     vector<bool> zusj(alf);  
     double probsum=0.0;
     whichwildcard(j,zusj);
     for(int k=0;k<alf;++k)
	 if(zusj[k])
	    probsum+=probs_time[i][k];
     return log(probsum);
  } 
  if(j<alf) {
     vector<bool> zusi(alf);  
     double probsum=0.0;
     whichwildcard(i,zusi);
     for(int k=0;k<alf;++k)
	 if(zusi[k])
	    probsum+=statefreqs[k]*probs_time[k][j];
     return log(probsum)-logfreqofstate(i);
  }
  vector<bool> zusi(alf);  
  vector<bool> zusj(alf);  
  double probsum=0.0, probsum2;
  whichwildcard(i,zusi);
  whichwildcard(j,zusj);
  for(int k=0;k<alf;++k) 
      if(zusi[k]) {
	 probsum2=0.0;
	 for(int l=0;l<alf;++l)
	     if(zusj[l])
		probsum2+=probs_time[k][l];
	 probsum+=statefreqs[zusi[k]]*probsum2;
      }
  return log(probsum)-logfreqofstate(i);
}

double substmodel::logProb_ij(int i, int j, double ptime)
{
   /* gibt den i,j-ten Eintrag in der logW'keitsmatrix
     zurueck */ 
  if(time > ptime+DBL_MIN || time < ptime-DBL_MIN)
    settime(ptime);

  if(i<alf && j<alf)
     return log(probs_time[i][j]);
  if(i<alf) {
     vector<bool> zusj(alf);  
     double probsum=0.0;
     whichwildcard(j,zusj);
     for(int k=0;k<alf;++k)
	 if(zusj[k])
	    probsum+=probs_time[i][k];
     return log(probsum);
  } 
  if(j<alf) {
     vector<bool> zusi(alf);  
     double probsum=0.0;
     whichwildcard(i,zusi);
     for(int k=0;k<alf;++k)
	 if(zusi[k])
	    probsum+=statefreqs[k]*probs_time[k][j];
     return log(probsum)-logfreqofstate(i);
  }
  vector<bool> zusi(alf);  
  vector<bool> zusj(alf);  
  double probsum=0.0, probsum2;
  whichwildcard(i,zusi);
  whichwildcard(j,zusj);
  for(int k=0;k<alf;++k) 
      if(zusi[k]) {
	 probsum2=0.0;
	 for(int l=0;l<alf;++l)
	     if(zusj[l])
		probsum2+=probs_time[k][l];
	 probsum+=statefreqs[zusi[k]]*probsum2;
      }
  return log(probsum)-logfreqofstate(i);
}

/*
double substmodel::Prob_ij(int i, int j)
{
   gibt den i,j-ten Eintrag in der W'keitsmatrix
     zurueck 
  if(i>=alf || j>=alf)
  {
    cerr << "Indices too big: i = " << i << ", j = " << j << "\n";
    exit(EXIT_FAILURE);
  }

  return probs_time[i][j];
}

double substmodel::Prob_ij(int i, int j, double ptime)
{
   gibt den i,j-ten Eintrag in der W'keitsmatrix
     zurueck 
  if(time > ptime+DBL_MIN || time < ptime-DBL_MIN)
    settime(ptime);

  return Prob_ij(i,j);
}

double substmodel::Prob_cd(char c, char d)
{
   gibt die W'keit des Paares (c,d)
     zurueck, d.h. die Haeufigkeit
     von c ist hier schon beruecksichtigt.
     Ausserdem kann diese Funktion auch
     nicht eindeutig bestimmte Basen bzw AmAcs verarbeiten
  double probCD;
  
  switch (alf) {
     case DNA :
       probCD = NucProb_cd(c,d);
       break;
     case PROTEIN :
       probCD = AAProb_cd(c,d);
       break;
     default :
       cerr << "The program can only deal with nucleotide and\n"
            << "amino acid sequences.\n";
       exit(EXIT_FAILURE);
  }
  
  return probCD;
}


double substmodel::Prob_cd(char c, char d, double ptime)
{
   gibt die W'keit des Paares (c,d)
     zurueck, d.h. die Haeufigkeit
     von c ist hier schon beruecksichtigt.
     Ausserdem kann diese Funktion auch
     nicht eindeutig bestimmte Basen bzw AmAcs verarbeiten
  if(time > ptime+DBL_MIN || time < ptime-DBL_MIN)
    settime(ptime);

  return Prob_cd(c,d);
}


double substmodel::NucProb_cd(char c, char d)
{
   gibt die W'keit des Paares (c,d)
     zurueck, d.h. die Basenhaeufigkeit
     von c ist hier schon beruecksichtigt.
     Ausserdem kann diese Funktion auch
     nicht eindeutig bestimmte Basen verarbeiten
  double probCD;
  unsigned int i,j;
  vector<int> anc, rec;
  vector<double> transto;
  
  if(c=='N' && d=='N')
     probCD=1.0;
  else {
     chartobase(c,anc);
     chartobase(d,rec);
  
     probCD=0.0;
     for(i=0; i<anc.size(); ++i) {
       transto.push_back(0.0);
       for(j=0;j<rec.size();++j)
          transto[i]+=probs_time[anc[i]][rec[j]];
       probCD+=statefreqs[anc[i]]*transto[i];
     }
  }

  return probCD;
}

double substmodel::AAProb_cd(char c, char d)
{
   gibt die W'keit des Aminosaeurepaares (c,d)
     zurueck, d.h. die Haeufigkeit
     von c ist hier schon beruecksichtigt.
     Ausserdem kann diese Funktion auch
     nicht eindeutig bestimmte Aminosaeuren verarbeiten
  double probCD;
  vector<int> anc, rec;
  vector<double> transto;
  unsigned int i,j;
  
  if(c=='X' && d=='X')
     probCD=1.0;
  else {
     chartoamac(c,anc);
     chartoamac(d,rec);
  
     probCD=0.0;
     for(i=0; i<anc.size(); ++i) {
       transto.push_back(0.0);
       for(j=0;j<rec.size();++j)
          transto[i]+=probs_time[anc[i]][rec[j]];
       probCD+=statefreqs[anc[i]]*transto[i];
     }
  }

  return probCD;
}

double substmodel::Erw_Prob_prof(const vector<double>& prof1, const vector<double>& prof2)
{
    double sumj, propro=0.0;
    int i, j;
    
    if(prof1.size()!=prof2.size() || (int)prof1.size()!=alf) {
       cerr << "Error in function Prob_prof "<<prof1.size()<<" "<< prof2.size()<<" "<<(int)alf<<endl;
       exit(EXIT_FAILURE);
    }
    for(i=0;i<alf;++i) 
      if(prof1[i]>=DBL_MIN) {
        sumj = 0.0;
        for(j=0;j<alf;++j)
	  if(prof2[j]>=DBL_MIN)
            sumj+=probs_time[i][j]*prof2[j];
	if(sumj>-DBL_MIN && sumj<DBL_MIN) {
	   cerr << "Second profile empty in function Erw_Prob_prof\n";
	   exit(EXIT_FAILURE);
	}
        propro+=prof1[i]*statefreqs[i]*sumj;
    }
    if(propro>-DBL_MIN && propro<DBL_MIN) {
	cerr << "First profile empty in function Erw_Prob_prof\n";
	exit(EXIT_FAILURE);
    }

    return propro;
}


double substmodel::Erw_Prob_prof(const vector<double>& prof1, const vector<double>& prof2, double ptime)
{
   des Profilpaares prof1 und prof2 wieder 
  if(time > ptime+DBL_MIN || time < ptime-DBL_MIN)
    settime(ptime);

  return Erw_Prob_prof(prof1,prof2);
}*/


void substmodel::onepamratematrix()
// rate matrix with 0.01 substitutions per time unit
{

   double delta, temp, sum;
   vector<double> m;
   int i, j;
   

   if(alf!=(int)statefreqs.size()) {
      cerr << "Alphabet has size " << (int)alf <<",\n"
           << "but the distribution has  " << (int)statefreqs.size() << " states!\n";
      exit(EXIT_FAILURE);  
   }   
   if(alf!=(int)rates.size()) {
      rates.resize(alf);
      for(i=0;i<(int)rates.size();++i)
         rates[i].resize(alf);
   }
   
   // fill Q*
   switch(alf) {
      case DNA :
         switch (model) {
	   case GENREV :
	      genrevmod();
              break;
	   case TAMNEI :
              tamneimod();
	      break;
	   case HKY :
              hkymod();
	      break;
	   case KIMURA2P :
              kimura2mod();
	      break;
	   case KIMURA3P :
              kimura3mod();
	      break;
	   case FEL81 :
              fel81mod();
	      break;
	   case JC69 :
              jcmod();
	      break;
	   default :
	      cerr << "You were trying to analyse nucleotide sequences\n"
	           << "with an inadequate substitution model.\n";
	      exit(EXIT_FAILURE);
	 }
         break;
      case PROTEIN :
         switch (model) {
	   case MV2000 :
	      mv2000mod();
	      break;
	   case JTT :
	      jttdata();
	      break;
	   case DAYHOFF :
	      dyhfdata();
	      break;
	   case MTREV24 :
	      mtrevdata();
	      break;
	   case BLOSUM62 :
	      blosum62data();
	      break;
	   default :
	      cerr << "You were trying to analyse amino acid sequences\n"
	           << "with an inadequate substitution model.\n";
	      exit(EXIT_FAILURE);
	 }
	 break;
      default :
         cerr << "Unknown sequence type with " <<(int)alf <<" states\n";
         exit(EXIT_FAILURE);
   }
   
   
   //1Pam
   for (i = 0; i < alf; ++i)
   {
     for (j = 0; j < alf; ++j)
     {
       rates[i][j] = statefreqs[j]*rates[i][j];
     }
   }

   m.resize(alf);
   for (i = 0, sum = 0.0; i < alf; ++i)
   {
     for (j = 0, temp = 0.0; j < alf; ++j)
	temp += rates[i][j];
     m[i] = temp; /* row sum */
     sum += temp*statefreqs[i]; /* exp. rate */
   }
   delta = 0.01 / sum; /* 0.01 subst. per unit time */
   for (i = 0; i < alf; ++i) {
     for (j = 0; j < alf; ++j) {
         if (i != j)
            rates[i][j] = delta * rates[i][j];
         else
            rates[i][j] = delta * (-m[i]);
     }
   }

}


void substmodel::tranprobmat()
/* compute 1 PAM rate matrix, its eigensystem, and the inverse matrix thereof */
{
   eigensystem(); /* eigensystem of 1 PAM rate matrix */
   luinverse(); /* inverse eigenvectors are in Ievc */
   checkevector(); /* check whether inversion was OK */
   //fillogstatefreqs();
}


/* compute P(t) and logprobs_time*/
void substmodel::tprobmtrx()
{
    register int i, j, k;
    register double temp;

    for (k = 0; k < alf; ++k) {
    	temp = exp(time * Eval[k]);
    	for (j = 0; j < alf; ++j)
    		iexp[k][j] = Ievc[k][j] * temp;
    }
    for (i = 0; i < alf; ++i) {
    	for (j = 0; j < alf; ++j) {
    		temp = 0.0;
    		for (k = 0; k < alf; ++k)
    			temp += Evec[i][k] * iexp[k][j];
   		probs_time[i][j] = fabs(temp);
    	}
    }
    //fillogprobs();
}

/*
void substmodel::fillogprobs()
{
    register int i, j;
    unsigned int k,l;
    bool bdummy=FALSE;
    vector<bool> zusi(alf), zusj(alf);
    double probsum=0.0, probsum2=0.0;
    
    for(i=0;i<alf;++i) {
        for(j=0;j<alf;++j)
	    logprobs_time[i][j]=log(probs_time[i][j]);
	for(j=alf;j<logalf;++j) {
            if((j==dNtp && alf==DNA)||(j==XXX && alf==PROTEIN)) {
                logprobs_time[i][j]=0.0;
            } else {
	        probsum=0.0;
                whichwildcard(j,zusj);
		for(k=0;k<alf;++k)
		   if(zusj[k])
		      probsum+=probs_time[i][k];
		logprobs_time[i][j]=log(probsum);
	    }
	}
    }
    for(i=alf;i<logalf;++i) {
        for(j=0;j<alf;++j) {
            probsum=0.0;
            whichwildcard(i,zusi);
	    for(k=0;k<alf;++k)
		if(zusi[k])
		   probsum+=statefreqs[k]*probs_time[k][j];
            logprobs_time[i][j]=log(probsum)-logstatefreqs[i];
	}
	for(j=alf;j<logalf;++j) {
            if((j==dNtp && alf==DNA)||(j==XXX && alf==PROTEIN)) {
                logprobs_time[i][j]=0.0;
            } else {
	        probsum=0.0;
                whichwildcard(i,zusi);
                whichwildcard(j,zusj);
	        for(k=0;k<alf;++k) 
		   if(zusi[k]) {
	            probsum2=0.0;
		    for(l=0;l<alf;++l)
		       if(zusj[l])
		          probsum2+=probs_time[k][l];
	            probsum+=statefreqs[zusi[k]]*probsum2;
	        }
                logprobs_time[i][j]=log(probsum)-logstatefreqs[i];
	    }
	}
    }
}

void substmodel::fillogstatefreqs()
{
   int i;
   unsigned int j;
   double probsum=0.0;
   vector<bool> zus(alf);
   
   for(i=0;i<alf;++i)
       logstatefreqs[i]=log(statefreqs[i]);
   for(i=alf;i<logalf;++i) {
       if((i==dNtp && alf==DNA)||(i==XXX && alf==PROTEIN)) {
           logstatefreqs[i]=0.0;
       } else {
           probsum=0.0;
           whichwildcard(i,zus);
           for(j=0;j<alf;++j)
	      if(zus[j])
                 probsum+=statefreqs[j];
           logstatefreqs[i]=log(probsum);
       }
   }
}
*/
/***********************************************************/
/***** functions to compute the eigensystem ****************/
/***********************************************************/

void substmodel::eigensystem()
{
   dvector evali;
   dmatrix b;
   ivector ordr;
   int i, j, k, error, tpmradix;
   double zero;
   
   tpmradix = alf;
   evali = new_dvector(tpmradix);
   b = new_dmatrix(tpmradix,tpmradix);
   ordr = new_ivector(tpmradix);

   onepamratematrix(); /* make 1 PAM rate matrix */
	
   /* copy a to b */
   for (i = 0; i < tpmradix; ++i)
     for (j = 0; j < tpmradix; ++j)
	 b[i][j] = rates[i][j];

   elmhes(b, ordr, tpmradix); /* compute eigenvalues and eigenvectors */
   eltran(b, ordr, tpmradix);
   hqr2(tpmradix, 1, tpmradix, b, evali);
	
   /* check eigenvalue equation */
   error = FALSE;
   for (j = 0; j < tpmradix; ++j) {
     for (i = 0, zero = 0.0; i < tpmradix; ++i) {
       for (k = 0; k < tpmradix; ++k) zero += rates[i][k] * Evec[k][j];
          zero -= Eval[j] * Evec[i][j];
          if (fabs(zero) > 1.0e-5)
               error = TRUE;	
     }
   }
   if (error)
      cerr << "\nWARNING: Eigensystem doesn't satisfy eigenvalue equation!\n";
      
   free_dvector(evali);
   free_dmatrix(b);
   free_ivector(ordr);

}

void substmodel::luinverse()
{
    double eps = 1.0e-20; /* ! */
    int i, j, k, l, maxi=0, idx, ix, jx, tpmradix;
    double sum, tmp, maxb, aw;
    ivector index;
    double *wk;
    dmatrix omtrx;

    tpmradix = alf;
    index = new_ivector(tpmradix);
    omtrx = new_dmatrix(tpmradix,tpmradix);
    
    /* copy Evec to omtrx */
    for (i = 0; i < tpmradix; ++i)
    	for (j = 0; j < tpmradix; ++j)
    		 omtrx[i][j] = Evec[i][j];
	
    wk = (double *) malloc((unsigned)tpmradix * sizeof(double));
    aw = 1.0;
    for (i = 0; i < tpmradix; ++i) {
    	maxb = 0.0;
    	for (j = 0; j < tpmradix; ++j) {
    		if (fabs(omtrx[i][j]) > maxb)
    			maxb = fabs(omtrx[i][j]);
    	}
    	if (maxb == 0.0) {
    		/* Singular matrix */
    		cerr << "\n\n\nHALT: PLEASE REPORT ERROR C TO DEVELOPERS\n\n\n";
    		exit(1);
    	}
    	wk[i] = 1.0 / maxb;
    }
    for (j = 0; j < tpmradix; ++j) {
    	for (i = 0; i < j; ++i) {
    		sum = omtrx[i][j];
    		for (k = 0; k < i; ++k)
    			sum -= omtrx[i][k] * omtrx[k][j];
    		omtrx[i][j] = sum;
    	}
    	maxb = 0.0;
    	for (i = j; i < tpmradix; ++i) {
    		sum = omtrx[i][j];
    		for (k = 0; k < j; ++k)
    			sum -= omtrx[i][k] * omtrx[k][j];
    		omtrx[i][j] = sum;
    		tmp = wk[i] * fabs(sum);
    		if (tmp >= maxb) {
    			maxb = tmp;
    			maxi = i;
    		}
    	}
    	if (j != maxi) {
    		for (k = 0; k < tpmradix; ++k) {
    			tmp = omtrx[maxi][k];
    			omtrx[maxi][k] = omtrx[j][k];
    			omtrx[j][k] = tmp;
    		}
    		aw = -aw;
    		wk[maxi] = wk[j];
    	}
    	index[j] = maxi;
    	if (omtrx[j][j] == 0.0)
    		omtrx[j][j] = eps;
    	if (j != tpmradix - 1) {
    		tmp = 1.0 / omtrx[j][j];
    		for (i = j + 1; i < tpmradix; ++i)
    			omtrx[i][j] *= tmp;
    	}
    }
    for (jx = 0; jx < tpmradix; ++jx) {
    	for (ix = 0; ix < tpmradix; ++ix)
    		wk[ix] = 0.0;
    	wk[jx] = 1.0;
    	l = -1;
    	for (i = 0; i < tpmradix; ++i) {
    		idx = index[i];
    		sum = wk[idx];
    		wk[idx] = wk[i];
    		if (l != -1) {
    			for (j = l; j < i; ++j)
    				sum -= omtrx[i][j] * wk[j];
    		} else if (sum != 0.0)
    			l = i;
    		wk[i] = sum;
    	}
    	for (i = tpmradix - 1; i >= 0; --i) {
    		sum = wk[i];
    		for (j = i + 1; j < tpmradix; ++j)
    			sum -= omtrx[i][j] * wk[j];
    		wk[i] = sum / omtrx[i][i];
    	}
    	for (ix = 0; ix < tpmradix; ++ix)
    		Ievc[ix][jx] = wk[ix];
    }
    free((char *)wk);
    free_ivector(index);
    free_dmatrix(omtrx);
    wk = NULL;
}


void substmodel::checkevector()
{
	int i, j, ia, ib, ic, error, nn;
	dmatrix matx;
	double sum;

        nn = alf;
        matx = new_dmatrix(nn,nn);
	
	/* multiply matrix of eigenvectors and its inverse */
	for (ia = 0; ia < nn; ++ia) {
		for (ic = 0; ic < nn; ++ic) {
			sum = 0.0;
			for (ib = 0; ib < nn; ++ib) sum += Evec[ia][ib] * Ievc[ib][ic];
			matx[ia][ic] = sum;
		}
	}
	/* check whether the unitary matrix is obtained */
	error = FALSE;
	for (i = 0; i < nn; ++i) {
		for (j = 0; j < nn; ++j) {
			if (i == j) {
				if (fabs(matx[i][j] - 1.0) > 1.0e-5)
					error = TRUE;
			} else {
				if (fabs(matx[i][j]) > 1.0e-5)
					error = TRUE;
			}
		}
	}
	if (error) {
		cerr << "\nWARNING: Inversion of eigenvector matrix not perfect!\n";
	}
	
	free_dmatrix(matx);
}

/***************************** internal functions *****************************/


void substmodel::elmhes(double **b, int *ordr, int n)
{
	int m, j, i;
	double y, x;


	for (i = 0; i < n; ++i)
		ordr[i] = 0;
	for (m = 2; m < n; ++m) {
		x = 0.0;
		i = m;
		for (j = m; j <= n; ++j) {
			if (fabs(b[j - 1][m - 2]) > fabs(x)) {
				x = b[j - 1][m - 2];
				i = j;
			}
		}
		ordr[m - 1] = i;      /* vector */
		if (i != m) {
			for (j = m - 2; j < n; ++j) {
				y = b[i - 1][j];
				b[i - 1][j] = b[m - 1][j];
				b[m - 1][j] = y;
			}
			for (j = 0; j < n; ++j) {
				y = b[j][i - 1];
				b[j][i - 1] = b[j][m - 1];
				b[j][m - 1] = y;
			}
		}
		if (x != 0.0) {
			for (i = m; i < n; ++i) {
				y = b[i][m - 2];
				if (y != 0.0) {
					y /= x;
					b[i][m - 2] = y;
					for (j = m - 1; j < n; ++j)
						b[i][j] -= y * b[m - 1][j];
					for (j = 0; j < n; ++j)
						b[j][m - 1] += y * b[j][i];
				}
			}
		}
	}
}


void substmodel::eltran(double **b, int *ordr, int n)
{
	int i, j, m;


	for (i = 0; i < n; ++i) {
		for (j = i + 1; j < n; ++j) {
			Evec[i][j] = 0.0;
			Evec[j][i] = 0.0;
		}
		Evec[i][i] = 1.0;
	}
	if (n <= 2)
		return;
	for (m = n - 1; m >= 2; --m) {
		for (i = m; i < n; ++i)
			Evec[i][m - 1] = b[i][m - 2];
		i = ordr[m - 1];
		if (i != m) {
			for (j = m - 1; j < n; ++j) {
				Evec[m - 1][j] = Evec[i - 1][j];
				Evec[i - 1][j] = 0.0;
			}
			Evec[i - 1][m - 1] = 1.0;
		}
	}
}


void substmodel::mcdiv(double ar, double ai, double br, double bi,
	double *cr, double *ci)
{
	double s, ars, ais, brs, bis;


	s = fabs(br) + fabs(bi);
	ars = ar / s;
	ais = ai / s;
	brs = br / s;
	bis = bi / s;
	s = brs * brs + bis * bis;
	*cr = (ars * brs + ais * bis) / s;
	*ci = (ais * brs - ars * bis) / s;
}


void substmodel::hqr2(int n, int low, int hgh, double **b, double *wi)
{
	int i, j, k, l=0, m, en, na, itn, its;
	double p=0, q=0, r=0, s=0, t, w, x=0, y, ra, sa, vi, vr, z=0, norm, tst1, tst2;
	int notlas; /* boolean */


	norm = 0.0;
	k = 1;
	/* store isolated roots and compute matrix norm */
	for (i = 0; i < n; ++i) {
		for (j = k - 1; j < n; ++j)
			norm += fabs(b[i][j]);
		k = i + 1;
		if (i + 1 < low || i + 1 > hgh) {
			Eval[i] = b[i][i];
			wi[i] = 0.0;
		}
	}
	en = hgh;
	t = 0.0;
	itn = n * 1000;
	while (en >= low) {	/* search for next eigenvalues */
		its = 0;
		na = en - 1;
		while (en >= 1) {
			/* look for single small sub-diagonal element */
			for (l = en; l > low; --l) {
				s = fabs(b[l - 2][l - 2]) + fabs(b[l - 1][l - 1]);
				if (s == 0.0)
					s = norm;
				tst1 = s;
				tst2 = tst1 + fabs(b[l - 1][l - 2]);
				if (tst2 == tst1)
					goto L100;
			}
			l = low;
	L100:
			x = b[en - 1][en - 1];	/* form shift */
			if (l == en || l == na)
				break;
			if (itn == 0) {
				/* all eigenvalues have not converged */
				cerr << "\n\n\nHALT: PLEASE REPORT ERROR B TO DEVELOPERS\n\n\n";
				exit(1);
			}
			y = b[na - 1][na - 1];
			w = b[en - 1][na - 1] * b[na - 1][en - 1];
			/* form exceptional shift */
			if (its == 10 || its == 20) {
				t += x;
				for (i = low - 1; i < en; ++i)
					b[i][i] -= x;
				s = fabs(b[en - 1][na - 1]) + fabs(b[na - 1][en - 3]);
				x = 0.75 * s;
				y = x;
				w = -0.4375 * s * s;
			}
			++its;
			--itn;
			/* look for two consecutive small sub-diagonal elements */
			for (m = en - 2; m >= l; --m) {
				z = b[m - 1][m - 1];
				r = x - z;
				s = y - z;
				p = (r * s - w) / b[m][m - 1] + b[m - 1][m];
				q = b[m][m] - z - r - s;
				r = b[m + 1][m];
				s = fabs(p) + fabs(q) + fabs(r);
				p /= s;
				q /= s;
				r /= s;
				if (m == l)
					break;
				tst1 = fabs(p) *
						(fabs(b[m - 2][m - 2]) + fabs(z) + fabs(b[m][m]));
				tst2 = tst1 + fabs(b[m - 1][m - 2]) * (fabs(q) + fabs(r));
				if (tst2 == tst1)
					break;
			}
			for (i = m + 2; i <= en; ++i) {
				b[i - 1][i - 3] = 0.0;
				if (i != m + 2)
					b[i - 1][i - 4] = 0.0;
			}
			for (k = m; k <= na; ++k) {
				notlas = (k != na);
				if (k != m) {
					p = b[k - 1][k - 2];
					q = b[k][k - 2];
					r = 0.0;
					if (notlas)
						r = b[k + 1][k - 2];
					x = fabs(p) + fabs(q) + fabs(r);
					if (x != 0.0) {
						p /= x;
						q /= x;
						r /= x;
					}
				}
				if (x != 0.0) {
					if (p < 0.0) /* sign */
						s = - sqrt(p * p + q * q + r * r);
					else
						s = sqrt(p * p + q * q + r * r);
					if (k != m)
						b[k - 1][k - 2] = -s * x;
					else {
						if (l != m)
							b[k - 1][k - 2] = -b[k - 1][k - 2];
					}
					p += s;
					x = p / s;
					y = q / s;
					z = r / s;
					q /= p;
					r /= p;
					if (!notlas) {
						for (j = k - 1; j < n; ++j) {	/* row modification */
							p = b[k - 1][j] + q * b[k][j];
							b[k - 1][j] -= p * x;
							b[k][j] -= p * y;
						}
						j = (en < (k + 3)) ? en : (k + 3); /* min */
						for (i = 0; i < j; ++i) {	/* column modification */
							p = x * b[i][k - 1] + y * b[i][k];
							b[i][k - 1] -= p;
							b[i][k] -= p * q;
						}
						/* accumulate transformations */
						for (i = low - 1; i < hgh; ++i) {
							p = x * Evec[i][k - 1] + y * Evec[i][k];
							Evec[i][k - 1] -= p;
							Evec[i][k] -= p * q;
						}
					} else {
						for (j = k - 1; j < n; ++j) {	/* row modification */
							p = b[k - 1][j] + q * b[k][j] + r * b[k + 1][j];
							b[k - 1][j] -= p * x;
							b[k][j] -= p * y;
							b[k + 1][j] -= p * z;
						}
						j = (en < (k + 3)) ? en : (k + 3); /* min */
						for (i = 0; i < j; ++i) {	/* column modification */
							p = x * b[i][k - 1] + y * b[i][k] + z * b[i][k + 1];
							b[i][k - 1] -= p;
							b[i][k] -= p * q;
							b[i][k + 1] -= p * r;
						}
						/* accumulate transformations */
						for (i = low - 1; i < hgh; ++i) {
							p = x * Evec[i][k - 1] + y * Evec[i][k] +
								z * Evec[i][k + 1];
							Evec[i][k - 1] -= p;
							Evec[i][k] -= p * q;
							Evec[i][k + 1] -= p * r;
						}
					}
				}
			}	       /* for k */
		}		       /* while infinite loop */
		if (l == en) {	       /* one root found */
			b[en - 1][en - 1] = x + t;
			Eval[en - 1] = b[en - 1][en - 1];
			wi[en - 1] = 0.0;
			en = na;
			continue;
		}
		y = b[na - 1][na - 1];
		w = b[en - 1][na - 1] * b[na - 1][en - 1];
		p = (y - x) / 2.0;
		q = p * p + w;
		z = sqrt(fabs(q));
		b[en - 1][en - 1] = x + t;
		x = b[en - 1][en - 1];
		b[na - 1][na - 1] = y + t;
		if (q >= 0.0) {	       /* real pair */
			if (p < 0.0) /* sign */
				z = p - fabs(z);
			else
				z = p + fabs(z);
			Eval[na - 1] = x + z;
			Eval[en - 1] = Eval[na - 1];
			if (z != 0.0)
				Eval[en - 1] = x - w / z;
			wi[na - 1] = 0.0;
			wi[en - 1] = 0.0;
			x = b[en - 1][na - 1];
			s = fabs(x) + fabs(z);
			p = x / s;
			q = z / s;
			r = sqrt(p * p + q * q);
			p /= r;
			q /= r;
			for (j = na - 1; j < n; ++j) {	/* row modification */
				z = b[na - 1][j];
				b[na - 1][j] = q * z + p * b[en - 1][j];
				b[en - 1][j] = q * b[en - 1][j] - p * z;
			}
			for (i = 0; i < en; ++i) {	/* column modification */
				z = b[i][na - 1];
				b[i][na - 1] = q * z + p * b[i][en - 1];
				b[i][en - 1] = q * b[i][en - 1] - p * z;
			}
			/* accumulate transformations */
			for (i = low - 1; i < hgh; ++i) {
				z = Evec[i][na - 1];
				Evec[i][na - 1] = q * z + p * Evec[i][en - 1];
				Evec[i][en - 1] = q * Evec[i][en - 1] - p * z;
			}
		} else {	       /* complex pair */
			Eval[na - 1] = x + p;
			Eval[en - 1] = x + p;
			wi[na - 1] = z;
			wi[en - 1] = -z;
		}
		en -= 2;
	}			       /* while en >= low */
	/* backsubstitute to find vectors of upper triangular form */
	if (norm != 0.0) {
		for (en = n; en >= 1; --en) {
			p = Eval[en - 1];
			q = wi[en - 1];
			na = en - 1;
			if (q == 0.0) {/* real vector */
				m = en;
				b[en - 1][en - 1] = 1.0;
				if (na != 0) {
					for (i = en - 2; i >= 0; --i) {
						w = b[i][i] - p;
						r = 0.0;
						for (j = m - 1; j < en; ++j)
							r += b[i][j] * b[j][en - 1];
						if (wi[i] < 0.0) {
							z = w;
							s = r;
						} else {
							m = i + 1;
							if (wi[i] == 0.0) {
								t = w;
								if (t == 0.0) {
									tst1 = norm;
									t = tst1;
									do {
										t = 0.01 * t;
										tst2 = norm + t;
									} while (tst2 > tst1);
								}
								b[i][en - 1] = -(r / t);
							} else {	/* solve real equations */
								x = b[i][i + 1];
								y = b[i + 1][i];
								q = (Eval[i] - p) * (Eval[i] - p) + wi[i] * wi[i];
								t = (x * s - z * r) / q;
								b[i][en - 1] = t;
								if (fabs(x) > fabs(z))
									b[i + 1][en - 1] = (-r - w * t) / x;
								else
									b[i + 1][en - 1] = (-s - y * t) / z;
							}
							/* overflow control */
							t = fabs(b[i][en - 1]);
							if (t != 0.0) {
								tst1 = t;
								tst2 = tst1 + 1.0 / tst1;
								if (tst2 <= tst1) {
									for (j = i; j < en; ++j)
										b[j][en - 1] /= t;
								}
							}
						}
					}
				}
			} else if (q > 0.0) {
				m = na;
				if (fabs(b[en - 1][na - 1]) > fabs(b[na - 1][en - 1])) {
					b[na - 1][na - 1] = q / b[en - 1][na - 1];
					b[na - 1][en - 1] = (p - b[en - 1][en - 1]) /
														b[en - 1][na - 1];
				} else
					mcdiv(0.0, -b[na - 1][en - 1], b[na - 1][na - 1] - p, q,
						&b[na - 1][na - 1], &b[na - 1][en - 1]);
				b[en - 1][na - 1] = 0.0;
				b[en - 1][en - 1] = 1.0;
				if (en != 2) {
					for (i = en - 3; i >= 0; --i) {
						w = b[i][i] - p;
						ra = 0.0;
						sa = 0.0;
						for (j = m - 1; j < en; ++j) {
							ra += b[i][j] * b[j][na - 1];
							sa += b[i][j] * b[j][en - 1];
						}
						if (wi[i] < 0.0) {
							z = w;
							r = ra;
							s = sa;
						} else {
							m = i + 1;
							if (wi[i] == 0.0)
								mcdiv(-ra, -sa, w, q, &b[i][na - 1],
									&b[i][en - 1]);
							else {	/* solve complex equations */
								x = b[i][i + 1];
								y = b[i + 1][i];
								vr = (Eval[i] - p) * (Eval[i] - p);
								vr = vr + wi[i] * wi[i] - q * q;	
								vi = (Eval[i] - p) * 2.0 * q;
								if (vr == 0.0 && vi == 0.0) {
									tst1 = norm * (fabs(w) + fabs(q) + fabs(x) +
												   fabs(y) + fabs(z));
									vr = tst1;
									do {
										vr = 0.01 * vr;
										tst2 = tst1 + vr;
									} while (tst2 > tst1);
								}
								mcdiv(x * r - z * ra + q * sa,
									 x * s - z * sa - q * ra, vr, vi,
								     &b[i][na - 1], &b[i][en - 1]);
								if (fabs(x) > fabs(z) + fabs(q)) {
									b[i + 1]
										[na - 1] = (q * b[i][en - 1] -
													w * b[i][na - 1] - ra) / x;
									b[i + 1][en - 1] = (-sa - w * b[i][en - 1] -
														q * b[i][na - 1]) / x;
								} else
									mcdiv(-r - y * b[i][na - 1],
										 -s - y * b[i][en - 1], z, q,
									     &b[i + 1][na - 1], &b[i + 1][en - 1]);
							}
							/* overflow control */
							t = (fabs(b[i][na - 1]) > fabs(b[i][en - 1])) ?
								 fabs(b[i][na - 1]) : fabs(b[i][en - 1]);
							if (t != 0.0) {
								tst1 = t;
								tst2 = tst1 + 1.0 / tst1;
								if (tst2 <= tst1) {
									for (j = i; j < en; ++j) {
										b[j][na - 1] /= t;
										b[j][en - 1] /= t;
									}
								}
							}
						}
					}
				}
			}
		}
		/* end back substitution. vectors of isolated roots */
		for (i = 0; i < n; ++i) {
			if (i + 1 < low || i + 1 > hgh) {
				for (j = i; j < n; ++j)
					Evec[i][j] = b[i][j];
			}
		}
		/* multiply by transformation matrix to give vectors of
		 * original full matrix. */
		for (j = n - 1; j >= low - 1; --j) {
			m = ((j + 1) < hgh) ? (j + 1) : hgh; /* min */
			for (i = low - 1; i < hgh; ++i) {
				z = 0.0;
				for (k = low - 1; k < m; ++k)
					z += Evec[i][k] * b[k][j];
				Evec[i][j] = z;
			}
		}
	}
	return;
}


/************************************************************************************/
/****************       amino acid substitution models     **************************/
/************************************************************************************/

void substmodel::jttdata()
{
	/*
	 * JTT model for amino acid evolution
	 * D.T. Jones, W.R. Taylor, and J.M. Thornton
	 * "The rapid generation of mutation data matrices from protein sequences"
	 * CABIOS  vol. 8 no. 3 1992 pp. 275-282
	 */

	rates[ALA][ARG]=3.1628651460584e+00; 	rates[ALA][ASN]=3.2804935927860e+00;
	rates[ALA][ASP]=4.8477237048666e+00; 	rates[ALA][CYS]=3.4612244897959e+00;
	rates[ALA][GLN]=3.3130910900946e+00; 	rates[ALA][GLU]=6.3199473337722e+00;
	rates[ALA][GLY]=1.0440154440154e+01; 	rates[ALA][HIS]=1.3061224489796e+00;
	rates[ALA][ILE]=2.1726844583987e+00; 	rates[ALA][LEU]=1.8443597219107e+00;
	rates[ALA][LYS]=2.2137668626773e+00; 	rates[ALA][MET]=2.7210884353741e+00;
	rates[ALA][PHE]=8.3265306122449e-01; 	rates[ALA][PRO]=1.1537414965986e+01;
	rates[ALA][SER]=2.2838213546288e+01; 	rates[ALA][THR]=2.7007955724663e+01;
	rates[ALA][TRP]=5.1311953352770e-01; 	rates[ALA][TYR]=8.3673469387755e-01;
	rates[ALA][VAL]=1.7474335188621e+01; 
	
	rates[ARG][ASN]=2.6598918637222e+00; 	rates[ARG][ASP]=9.1014867485456e-01;
	rates[ARG][CYS]=6.1624649859944e+00; 	rates[ARG][GLN]=1.8036482885837e+01;
	rates[ARG][GLU]=1.8924731182796e+00; 	rates[ARG][GLY]=8.1810886516769e+00;
	rates[ARG][HIS]=1.9119717452198e+01; 	rates[ARG][ILE]=1.4410687351864e+00;
	rates[ARG][LEU]=2.2211961707760e+00; 	rates[ARG][LYS]=3.9239234676922e+01;
	rates[ARG][MET]=2.5060690943044e+00; 	rates[ARG][PHE]=3.9439775910364e-01;
	rates[ARG][PRO]=4.1953094963476e+00; 	rates[ARG][SER]=5.9016766126741e+00;
	rates[ARG][THR]=3.8437069743152e+00; 	rates[ARG][TRP]=7.6766706682673e+00;
	rates[ARG][TYR]=1.4173669467787e+00; 	rates[ARG][VAL]=1.0308123249300e+00;

	rates[ASN][ASP]=3.2226935854843e+01; 	rates[ASN][CYS]=1.8710963455150e+00;
	rates[ASN][GLN]=4.5351268130622e+00; 	rates[ASN][GLU]=3.3951344979102e+00;
	rates[ASN][GLY]=4.5987249708180e+00; 	rates[ASN][HIS]=2.3693774375271e+01;
	rates[ASN][ILE]=2.9235880398671e+00; 	rates[ASN][LEU]=8.0960899565551e-01;
	rates[ASN][LYS]=1.5024269384537e+01; 	rates[ASN][MET]=1.9003322259136e+00;
	rates[ASN][PHE]=4.3853820598007e-01; 	rates[ASN][PRO]=7.1083317047749e-01;
	rates[ASN][SER]=2.9456208772690e+01; 	rates[ASN][THR]=1.3735908553410e+01;
	rates[ASN][TRP]=1.6706217370669e-01; 	rates[ASN][TYR]=4.1661129568106e+00;
	rates[ASN][VAL]=9.7452934662237e-01; 
	
	rates[ASP][CYS]=6.2857142857143e-01; 	rates[ASP][GLN]=3.0662020905923e+00;
	rates[ASP][GLU]=4.5450549450549e+01; 	rates[ASP][GLY]=7.5402435402435e+00;
	rates[ASP][HIS]=6.0544672718586e+00; 	rates[ASP][ILE]=6.8808114961961e-01;
	rates[ASP][LEU]=3.6130902064968e-01; 	rates[ASP][LYS]=1.6718197057180e+00;
	rates[ASP][MET]=1.0879120879121e+00; 	rates[ASP][PHE]=1.9340659340659e-01;
	rates[ASP][PRO]=7.3949579831933e-01; 	rates[ASP][SER]=3.4196528109572e+00;
	rates[ASP][THR]=2.4749487800335e+00; 	rates[ASP][TRP]=3.4536891679749e-01;
	rates[ASP][TYR]=2.6895604395604e+00; 	rates[ASP][VAL]=1.8608058608059e+00;

	rates[CYS][GLN]=5.5191637630662e-01; 	rates[CYS][GLU]=3.2442396313364e-01;
	rates[CYS][GLY]=3.3297297297297e+00; 	rates[CYS][HIS]=4.3726708074534e+00;
	rates[CYS][ILE]=9.1868131868132e-01; 	rates[CYS][LEU]=9.9466248037677e-01;
	rates[CYS][LYS]=2.9830508474576e-01; 	rates[CYS][MET]=2.4095238095238e+00;
	rates[CYS][PHE]=4.1485714285714e+00; 	rates[CYS][PRO]=7.3949579831933e-01;
	rates[CYS][SER]=1.2862939958592e+01; 	rates[CYS][THR]=2.8125907990315e+00;
	rates[CYS][TRP]=6.8244897959184e+00; 	rates[CYS][TYR]=1.2885714285714e+01;
	rates[CYS][VAL]=3.7714285714286e+00; 
	
	rates[GLN][GLU]=2.0316061593796e+01; 	rates[GLN][GLY]=1.3922214897825e+00;
	rates[GLN][HIS]=3.3861536130889e+01; 	rates[GLN][ILE]=4.7172339855267e-01;
	rates[GLN][LEU]=4.2320327755868e+00; 	rates[GLN][LYS]=1.7835941652395e+01;
	rates[GLN][MET]=2.6573751451800e+00; 	rates[GLN][PHE]=2.7595818815331e-01;
	rates[GLN][PRO]=9.4992143198743e+00; 	rates[GLN][SER]=3.2350653941322e+00;
	rates[GLN][THR]=3.0973838067678e+00; 	rates[GLN][TRP]=1.0512692882031e+00;
	rates[GLN][TYR]=1.5331010452962e+00; 	rates[GLN][VAL]=1.0778164924506e+00;

	rates[GLU][GLY]=6.6857641051189e+00; 	rates[GLU][HIS]=1.4458024443999e+00;
	rates[GLU][ILE]=6.7068415455512e-01; 	rates[GLU][LEU]=5.7932850559579e-01;
	rates[GLU][LYS]=1.0365070686558e+01; 	rates[GLU][MET]=1.0138248847926e+00;
	rates[GLU][PHE]=2.6359447004608e-01; 	rates[GLU][PRO]=1.1291226167887e+00;
	rates[GLU][SER]=1.8337006611901e+00; 	rates[GLU][THR]=1.9520424900414e+00;
	rates[GLU][TRP]=6.9519420671494e-01; 	rates[GLU][TYR]=3.8018433179723e-01;
	rates[GLU][VAL]=2.7772657450077e+00; 
	
	rates[GLY][HIS]=1.2113479939567e+00; 	rates[GLY][ILE]=3.2670032670033e-01;
	rates[GLY][LEU]=4.1817641817642e-01; 	rates[GLY][LYS]=1.6354950592239e+00;
	rates[GLY][MET]=7.6447876447876e-01; 	rates[GLY][PHE]=3.0579150579151e-01;
	rates[GLY][PRO]=1.2391551215081e+00; 	rates[GLY][SER]=1.1138492529797e+01;
	rates[GLY][THR]=1.8888816176952e+00; 	rates[GLY][TRP]=3.3491450634308e+00;
	rates[GLY][TYR]=3.1853281853282e-01; 	rates[GLY][VAL]=2.8416988416988e+00;

	rates[HIS][ILE]=1.0931677018634e+00; 	rates[HIS][LEU]=3.2194389461470e+00;
	rates[HIS][LYS]=3.1498052426571e+00; 	rates[HIS][MET]=1.9130434782609e+00;
	rates[HIS][PHE]=2.7329192546584e+00; 	rates[HIS][PRO]=6.7304834977469e+00;
	rates[HIS][SER]=4.3726708074534e+00; 	rates[HIS][THR]=2.8162964522581e+00;
	rates[HIS][TRP]=7.8083407275954e-01; 	rates[HIS][TYR]=3.5118012422360e+01;
	rates[HIS][VAL]=7.2877846790890e-01; 
	
	rates[ILE][LEU]=1.4069798333535e+01; 	rates[ILE][LYS]=1.2292791953809e+00;
	rates[ILE][MET]=2.8366300366300e+01; 	rates[ILE][PHE]=4.7384615384615e+00;
	rates[ILE][PRO]=5.8780435251023e-01; 	rates[ILE][SER]=2.4105749323141e+00;
	rates[ILE][THR]=1.5243062022723e+01; 	rates[ILE][TRP]=8.2888540031397e-01;
	rates[ILE][TYR]=1.8434065934066e+00; 	rates[ILE][VAL]=5.7699633699634e+01;

	rates[LEU][LYS]=8.8039805231089e-01; 	rates[LEU][MET]=2.2425954997384e+01;
	rates[LEU][PHE]=1.5099529042386e+01; 	rates[LEU][PRO]=6.2626896912611e+00;
	rates[LEU][SER]=3.4917298022888e+00; 	rates[LEU][THR]=1.6109411169944e+00;
	rates[LEU][TRP]=3.2366001345593e+00; 	rates[LEU][TYR]=1.4505494505495e+00;
	rates[LEU][VAL]=1.0557823129252e+01; 
	
	rates[LYS][MET]=3.6577885391445e+00; 	rates[LYS][PHE]=1.4915254237288e-01;
	rates[LYS][PRO]=1.2868062479229e+00; 	rates[LYS][SER]=2.8162964522581e+00;
	rates[LYS][THR]=5.7494151926786e+00; 	rates[LYS][TRP]=5.4790729851263e-01;
	rates[LYS][TYR]=5.3268765133172e-01; 	rates[LYS][VAL]=7.4899112187248e-01;

	rates[MET][PHE]=2.5666666666667e+00; 	rates[MET][PRO]=9.4491129785247e-01;
	rates[MET][SER]=1.6397515527950e+00; 	rates[MET][THR]=1.2180790960452e+01;
	rates[MET][TRP]=1.1972789115646e+00; 	rates[MET][TYR]=1.1130952380952e+00;
	rates[MET][VAL]=1.7746031746032e+01; 
	
	rates[PHE][PRO]=8.8739495798319e-01; 	rates[PHE][SER]=5.6298136645963e+00;
	rates[PHE][THR]=8.3099273607748e-01; 	rates[PHE][TRP]=3.3224489795918e+00;
	rates[PHE][TYR]=3.3392857142857e+01; 	rates[PHE][VAL]=3.6000000000000e+00;

	rates[PRO][SER]=1.6261762676085e+01; 	rates[PRO][THR]=6.8852490148602e+00;
	rates[PRO][TRP]=4.2256902761104e-01; 	rates[PRO][TYR]=6.7787114845938e-01;
	rates[PRO][VAL]=1.2549019607843e+00; 
	
	rates[SER][THR]=2.7891216619293e+01; 	rates[SER][TRP]=1.8740017746229e+00;
	rates[SER][TYR]=3.7349896480331e+00; 	rates[SER][VAL]=2.4182194616977e+00;

	rates[THR][TRP]=4.8702870978900e-01; 	rates[THR][TYR]=1.1985472154964e+00;
	rates[THR][VAL]=6.7925746569814e+00; 
	
	rates[TRP][TYR]=4.6020408163265e+00; 	rates[TRP][VAL]=1.4693877551020e+00;

	rates[TYR][VAL]=1.0000000000000e+00; 


        // fill lower triangle of Q*
        for(int i=0;i<alf;++i) {
           for(int j=0;j<i;++j)
	       rates[i][j]=rates[j][i];
           rates[i][i]=0.0;  // Q* matrix!
        }
   
	statefreqs[ALA] = 0.077; statefreqs[ARG] = 0.051; statefreqs[ASN] = 0.043; statefreqs[ASP] = 0.052;
	statefreqs[CYS] = 0.02; statefreqs[GLN] = 0.041; statefreqs[GLU] = 0.062; statefreqs[GLY] = 0.074;
	statefreqs[HIS] = 0.023; statefreqs[ILE] = 0.052; statefreqs[LEU] = 0.091; statefreqs[LYS] = 0.059;
	statefreqs[MET] = 0.024; statefreqs[PHE] = 0.04; statefreqs[PRO] = 0.051; statefreqs[SER] = 0.069;
	statefreqs[THR] = 0.059; statefreqs[TRP] = 0.014; statefreqs[TYR] = 0.032; statefreqs[VAL] = 0.066;
}

void substmodel::dyhfdata()
{
	/*
	 * Dayhoff model for amino acid evolution
	 * Dayhoff, M.O., Schwartz, R.M., Orcutt, B.C. (1978)
	 * "A model of evolutionary change in proteins."
	 * Dayhoff, M.O. (ed.) Atlas of Protein Sequence Structur., Vol5, Suppl. 3,
	 * National Biomedical Research Foundation, Washington DC, pp. 345-352.
	 */

	rates[ALA][ARG]=9.6472567159749e-01; 	rates[ALA][ASN]=3.5927991886410e+00;
	rates[ALA][ASP]=4.3200552414656e+00; 	rates[ALA][CYS]=1.3184584178499e+00;
	rates[ALA][GLN]=3.2267534963169e+00; 	rates[ALA][GLU]=7.0141987829615e+00;
	rates[ALA][GLY]=8.5773867857875e+00; 	rates[ALA][HIS]=8.1434196396611e-01;
	rates[ALA][ILE]=2.3518447453539e+00; 	rates[ALA][LEU]=1.4735711728911e+00;
	rates[ALA][LYS]=9.3940162271805e-01; 	rates[ALA][MET]=2.5490196078431e+00;
	rates[ALA][PHE]=6.5922920892495e-01; 	rates[ALA][PRO]=8.9189834148670e+00;
	rates[ALA][SER]=1.4540712836859e+01; 	rates[ALA][THR]=1.3411904595370e+01;
	rates[ALA][TRP]=3.8517964118027e-02; 	rates[ALA][TYR]=8.7897227856660e-01;
	rates[ALA][VAL]=7.4036511156187e+00; 
	
	rates[ARG][ASN]=1.1890243902439e+00; 	rates[ARG][ASP]=5.9525626545377e-02;
	rates[ARG][CYS]=8.4778922655537e-01; 	rates[ARG][GLN]=8.8348561504191e+00;
	rates[ARG][GLU]=5.5954088952654e-02; 	rates[ARG][GLY]=3.1434881434075e-01;
	rates[ARG][HIS]=8.4753987678285e+00; 	rates[ARG][ILE]=2.2684090115941e+00;
	rates[ARG][LEU]=5.5954088952654e-01; 	rates[ARG][LYS]=1.6681312769010e+01;
	rates[ARG][MET]=3.1707317073171e+00; 	rates[ARG][PHE]=4.8959827833572e-01;
	rates[ARG][PRO]=3.6754156468900e+00; 	rates[ARG][SER]=5.4755072760812e+00;
	rates[ARG][THR]=9.6472567159749e-01; 	rates[ARG][TRP]=7.5538020086083e+00;
	rates[ARG][TYR]=2.7977044476327e-01; 	rates[ARG][VAL]=8.6083213773314e-01;

	rates[ASN][ASP]=3.2459324155194e+01; 	rates[ASN][CYS]=7.3852625416383e-02;
	rates[ASN][GLN]=3.7732198142415e+00; 	rates[ASN][GLU]=5.3911764705882e+00;
	rates[ASN][GLY]=5.0264375413087e+00; 	rates[ASN][HIS]=1.9061418685121e+01;
	rates[ASN][ILE]=2.7901430842607e+00; 	rates[ASN][LEU]=1.2482698961938e+00;
	rates[ASN][LYS]=1.1542279411765e+01; 	rates[ASN][MET]=1.9117647058824e-01;
	rates[ASN][PHE]=5.0183823529412e-01; 	rates[ASN][PRO]=1.5181660899654e+00;
	rates[ASN][SER]=1.7697478991597e+01; 	rates[ASN][THR]=8.3557302231237e+00;
	rates[ASN][TRP]=8.6029411764706e-01; 	rates[ASN][TYR]=3.4411764705882e+00;
	rates[ASN][VAL]=5.7352941176471e-01; 
	
	rates[ASP][CYS]=2.5534152404601e-02; 	rates[ASP][GLN]=4.8811013767209e+00;
	rates[ASP][GLU]=4.0561952440551e+01; 	rates[ASP][GLY]=4.4423506911730e+00;
	rates[ASP][HIS]=3.0865788117500e+00; 	rates[ASP][ILE]=8.5749078239692e-01;
	rates[ASP][LEU]=2.5926985518518e-02; 	rates[ASP][LYS]=2.5930851063830e+00;
	rates[ASP][MET]=1.1667143483333e-01; 	rates[ASP][PHE]=1.2963492759259e-02;
	rates[ASP][PRO]=4.7853935065891e-01; 	rates[ASP][SER]=3.4167709637046e+00;
	rates[ASP][THR]=2.3984722282163e+00; 	rates[ASP][TRP]=3.2408731898147e-02;
	rates[ASP][TYR]=8.1351689612015e-02; 	rates[ASP][VAL]=6.3829787234043e-01;

	rates[CYS][GLN]=2.1864264103535e-02; 	rates[CYS][GLU]=1.4770525083277e-02;
	rates[CYS][GLY]=3.9055458751427e-01; 	rates[CYS][HIS]=1.0223340673168e+00;
	rates[CYS][ILE]=1.5970515970516e+00; 	rates[CYS][LEU]=3.9098448749850e-02;
	rates[CYS][LYS]=8.0776309049169e-03; 	rates[CYS][MET]=1.4155086538140e-01;
	rates[CYS][PHE]=8.6898395721925e-02; 	rates[CYS][PRO]=6.8155604487784e-01;
	rates[CYS][SER]=5.8097784568373e+00; 	rates[CYS][THR]=5.9929928084086e-01;
	rates[CYS][TRP]=3.4759358288770e-01; 	rates[CYS][TYR]=3.4759358288770e+00;
	rates[CYS][VAL]=1.7647058823529e+00; 
	
	rates[GLN][GLU]=2.5476780185759e+01; 	rates[GLN][GLY]=1.0174974779977e+00;
	rates[GLN][HIS]=2.1573939173192e+01; 	rates[GLN][ILE]=6.5266504894988e-01;
	rates[GLN][LEU]=2.6634492806410e+00; 	rates[GLN][LYS]=5.5466331269350e+00;
	rates[GLN][MET]=4.0247678018576e+00; 	rates[GLN][PHE]=1.8038017885416e-02;
	rates[GLN][PRO]=5.5044618466582e+00; 	rates[GLN][SER]=2.0267580716497e+00;
	rates[GLN][THR]=1.9256432155439e+00; 	rates[GLN][TRP]=9.6202762055552e-02;
	rates[GLN][TYR]=1.0061919504644e-01; 	rates[GLN][VAL]=1.2538699690402e+00;

	rates[GLU][GLY]=2.8869795109055e+00; 	rates[GLU][HIS]=1.5519031141869e+00;
	rates[GLU][ILE]=2.1701112877583e+00; 	rates[GLU][LEU]=4.0484429065744e-01;
	rates[GLU][LYS]=2.9823529411765e+00; 	rates[GLU][MET]=1.0705882352941e+00;
	rates[GLU][PHE]=1.9801735189768e-02; 	rates[GLU][PRO]=1.7993079584775e+00;
	rates[GLU][SER]=2.8184873949580e+00; 	rates[GLU][THR]=1.2261663286004e+00;
	rates[GLU][TRP]=7.3114099162219e-02; 	rates[GLU][TYR]=7.6470588235294e-01;
	rates[GLU][VAL]=1.3058823529412e+00; 
	
	rates[GLY][HIS]=3.7906768788150e-01; 	rates[GLY][ILE]=2.3128004846840e-02;
	rates[GLY][LEU]=2.5776602775942e-01; 	rates[GLY][LYS]=9.6662260409782e-01;
	rates[GLY][MET]=6.0145406477198e-01; 	rates[GLY][PHE]=5.4775280898876e-01;
	rates[GLY][PRO]=1.2382877804129e+00; 	rates[GLY][SER]=8.2853366065527e+00;
	rates[GLY][THR]=1.1110604644803e+00; 	rates[GLY][TRP]=1.2888301387971e-01;
	rates[GLY][TYR]=1.7114723586662e-02; 	rates[GLY][VAL]=1.9233311302049e+00;

	rates[HIS][ILE]=2.7354343963341e-01; 	rates[HIS][LEU]=1.5876246692449e+00;
	rates[HIS][LYS]=9.6993944636678e-01; 	rates[HIS][MET]=1.2544085640577e-01;
	rates[HIS][PHE]=1.6868512110727e+00; 	rates[HIS][PRO]=3.3075513942601e+00;
	rates[HIS][SER]=1.2530894710826e+00; 	rates[HIS][THR]=8.1434196396611e-01;
	rates[HIS][TRP]=1.0121107266436e+00; 	rates[HIS][TYR]=4.4982698961938e+00;
	rates[HIS][VAL]=1.5570934256055e+00; 
	
	rates[ILE][LEU]=9.2275320303002e+00; 	rates[ILE][LYS]=1.6663354531002e+00;
	rates[ILE][MET]=1.1780604133545e+01; 	rates[ILE][PHE]=6.9753577106518e+00;
	rates[ILE][PRO]=4.2551201720752e-01; 	rates[ILE][SER]=8.8575970928912e-01;
	rates[ILE][THR]=6.8951811852420e+00; 	rates[ILE][TRP]=9.8802836705702e-02;
	rates[ILE][TYR]=1.3434022257552e+00; 	rates[ILE][VAL]=3.1526232114467e+01;

	rates[LEU][LYS]=6.5787197231834e-01; 	rates[LEU][MET]=1.8622837370242e+01;
	rates[LEU][PHE]=5.6340830449827e+00; 	rates[LEU][PRO]=1.1377976796255e+00;
	rates[LEU][SER]=6.1690558576372e-01; 	rates[LEU][THR]=1.2098794893211e+00;
	rates[LEU][TRP]=1.7543252595156e+00; 	rates[LEU][TYR]=1.0346020761246e+00;
	rates[LEU][VAL]=6.2906574394464e+00; 
	
	rates[LYS][MET]=8.6029411764706e+00; 	rates[LYS][PHE]=6.6640454965565e-03;
	rates[LYS][PRO]=1.2089100346021e+00; 	rates[LYS][SER]=3.4411764705882e+00;
	rates[LYS][THR]=4.9442190669371e+00; 	rates[LYS][TRP]=3.4272233982290e-02;
	rates[LYS][TYR]=4.7794117647059e-01; 	rates[LYS][VAL]=3.7500000000000e-01;

	rates[MET][PHE]=3.2500000000000e+00; 	rates[MET][PRO]=5.9976931949250e-01;
	rates[MET][SER]=2.1848739495798e+00; 	rates[MET][THR]=3.6916835699797e+00;
	rates[MET][TRP]=1.6247577591604e-01; 	rates[MET][TYR]=1.1508700794053e-01;
	rates[MET][VAL]=9.0588235294118e+00; 
	
	rates[PHE][PRO]=3.9359861591695e-01; 	rates[PHE][SER]=1.6386554621849e+00;
	rates[PHE][THR]=4.9442190669371e-01; 	rates[PHE][TRP]=2.8676470588235e+00;
	rates[PHE][TYR]=2.4852941176471e+01; 	rates[PHE][VAL]=4.4117647058824e-01;

	rates[PRO][SER]=8.6431043005437e+00; 	rates[PRO][THR]=2.8308077795013e+00;
	rates[PRO][TRP]=3.5840244687362e-02; 	rates[PRO][TYR]=4.3804743506776e-02;
	rates[PRO][VAL]=1.7301038062284e+00; 
	
	rates[SER][THR]=1.9663865546218e+01; 	rates[SER][TRP]=2.7857142857143e+00;
	rates[SER][TYR]=1.2016806722689e+00; 	rates[SER][VAL]=1.0840336134454e+00;

	rates[THR][TRP]=4.2019597219666e-02; 	rates[THR][TYR]=1.5162271805274e+00;
	rates[THR][VAL]=5.6592292089249e+00; 
	
	rates[TRP][TYR]=2.2941176470588e+00; 	rates[TRP][VAL]=1.2654363316538e-01;

	rates[TYR][VAL]=1.0000000000000e+00; 


        // fill lower triangle of Q*
        for(int i=0;i<alf;++i) {
           for(int j=0;j<i;++j)
	       rates[i][j]=rates[j][i];
           rates[i][i]=0.0;  // Q* matrix!
        }
   
	statefreqs[ALA] = 0.087; statefreqs[ARG] = 0.041; statefreqs[ASN] = 0.040; statefreqs[ASP] = 0.047;
	statefreqs[CYS] = 0.033; statefreqs[GLN] = 0.038; statefreqs[GLU] = 0.05; statefreqs[GLY] = 0.089;
	statefreqs[HIS] = 0.034; statefreqs[ILE] = 0.037; statefreqs[LEU] = 0.085; statefreqs[LYS] = 0.08;
	statefreqs[MET] = 0.015; statefreqs[PHE] = 0.04; statefreqs[PRO] = 0.051; statefreqs[SER] = 0.07;
	statefreqs[THR] = 0.058; statefreqs[TRP] = 0.01; statefreqs[TYR] = 0.03; statefreqs[VAL] = 0.065;
}

void substmodel::mtrevdata()
{	
	/*
	 * mtREV24 model of amino acid evolution
	 * (complete sequence data of mtDNA from 24 vertebrate species)
	 * Adachi, J. and Hasegawa, M. (1996)
	 */

	rates[ALA][ARG]=1.2199217606346e+01; 	rates[ALA][ASN]=1.4182139942122e+01;
	rates[ALA][ASP]=9.2985091873208e+00; 	rates[ALA][CYS]=3.1542792981957e+01;
	rates[ALA][GLN]=1.0025852846688e+00; 	rates[ALA][GLU]=5.1418866803338e+00;
	rates[ALA][GLY]=6.3531246495131e+01; 	rates[ALA][HIS]=7.3137132861715e+00;
	rates[ALA][ILE]=5.0782382656186e+01; 	rates[ALA][LEU]=1.3399741808481e+01;
	rates[ALA][LYS]=4.4021672780560e+00; 	rates[ALA][MET]=7.4673480520104e+01;
	rates[ALA][PHE]=3.3513021631978e+00; 	rates[ALA][PRO]=2.8582502221773e+01;
	rates[ALA][SER]=2.0413623195312e+02; 	rates[ALA][THR]=2.5301305153906e+02;
	rates[ALA][TRP]=1.0000000000000e+00; 	rates[ALA][TYR]=3.4084158197615e+00;
	rates[ALA][VAL]=1.0266468401249e+02; 
	
	rates[ARG][ASN]=6.9661274444534e+00; 	rates[ARG][ASP]=1.0000000000000e+00;
	rates[ARG][CYS]=5.4384584796568e+01; 	rates[ARG][GLN]=1.1631134513343e+02;
	rates[ARG][GLU]=1.0000000000000e+00; 	rates[ARG][GLY]=1.2122831341194e+01;
	rates[ARG][HIS]=8.6961067087353e+01; 	rates[ARG][ILE]=1.0000000000000e+00;
	rates[ARG][LEU]=8.1976829394538e+00; 	rates[ARG][LYS]=7.4423215395318e+01;
	rates[ARG][MET]=1.0000000000000e+00; 	rates[ARG][PHE]=2.4659158338099e+00;
	rates[ARG][PRO]=1.2439947713615e+01; 	rates[ARG][SER]=3.1791814866372e+00;
	rates[ARG][THR]=1.0935327216119e+00; 	rates[ARG][TRP]=1.1550775790126e+01;
	rates[ARG][TYR]=1.0000000000000e+00; 	rates[ARG][VAL]=4.0211417480338e+00;

	rates[ASN][ASP]=4.1809325468160e+02; 	rates[ASN][CYS]=3.1020979842967e+01;
	rates[ASN][GLN]=9.1349622725361e+01; 	rates[ASN][GLU]=3.3185663516310e+01;
	rates[ASN][GLY]=2.8052324651124e+01; 	rates[ASN][HIS]=2.6112087577885e+02;
	rates[ASN][ILE]=1.4261453863336e+01; 	rates[ASN][LEU]=7.9775653461977e+00;
	rates[ASN][LYS]=3.2036829276162e+02; 	rates[ASN][MET]=3.4424354918739e+01;
	rates[ASN][PHE]=7.9996445145608e+00; 	rates[ASN][PRO]=3.8586541461044e+01;
	rates[ASN][SER]=2.6020426225852e+02; 	rates[ASN][THR]=1.2550758780474e+02;
	rates[ASN][TRP]=5.6207759736659e+00; 	rates[ASN][TYR]=1.0071406219571e+02;
	rates[ASN][VAL]=1.0000000000000e+00; 
	
	rates[ASP][CYS]=1.0000000000000e+00; 	rates[ASP][GLN]=2.9097352675564e+01;
	rates[ASP][GLU]=3.0713149855302e+02; 	rates[ASP][GLY]=2.9877072751897e+01;
	rates[ASP][HIS]=5.9995408885817e+01; 	rates[ASP][ILE]=2.2827096245105e+00;
	rates[ASP][LEU]=1.0000000000000e+00; 	rates[ASP][LYS]=1.2183938185384e+00;
	rates[ASP][MET]=1.0000000000000e+00; 	rates[ASP][PHE]=2.6221929413096e+00;
	rates[ASP][PRO]=7.0708004204733e+00; 	rates[ASP][SER]=3.6327934317139e+01;
	rates[ASP][THR]=1.4743408713748e+01; 	rates[ASP][TRP]=1.0453246057102e+01;
	rates[ASP][TYR]=1.1165627147496e+01; 	rates[ASP][VAL]=1.0000000000000e+00;

	rates[CYS][GLN]=3.9599394038972e+01; 	rates[CYS][GLU]=1.0000000000000e+00;
	rates[CYS][GLY]=1.6163581056674e+01; 	rates[CYS][HIS]=7.4467985406234e+01;
	rates[CYS][ILE]=3.3018175376623e+01; 	rates[CYS][LEU]=1.3500725995091e+01;
	rates[CYS][LYS]=1.0000000000000e+00; 	rates[CYS][MET]=3.2504095376923e+00;
	rates[CYS][PHE]=3.7264767083096e+01; 	rates[CYS][PRO]=1.6454136037822e+01;
	rates[CYS][SER]=1.4581783243113e+02; 	rates[CYS][THR]=9.4720031458442e+01;
	rates[CYS][TRP]=1.7684087896962e+01; 	rates[CYS][TYR]=1.3409157685926e+02;
	rates[CYS][VAL]=1.0000000000000e+00; 
	
	rates[GLN][GLU]=1.6503249008836e+02; 	rates[GLN][GLY]=3.5530760735494e+00;
	rates[GLN][HIS]=3.0652523140859e+02; 	rates[GLN][ILE]=4.3905393139325e+00;
	rates[GLN][LEU]=2.0895470525345e+01; 	rates[GLN][LYS]=2.4504076430724e+02;
	rates[GLN][MET]=2.4931300477797e+01; 	rates[GLN][PHE]=1.0059428264289e+01;
	rates[GLN][PRO]=7.2256314165467e+01; 	rates[GLN][SER]=2.8480937892158e+01;
	rates[GLN][THR]=4.9962974409828e+01; 	rates[GLN][TRP]=1.0000000000000e+00;
	rates[GLN][TYR]=2.0430790980529e+01; 	rates[GLN][VAL]=9.9986289000676e+00;

	rates[GLU][GLY]=1.4884496769963e+01; 	rates[GLU][HIS]=2.5853576435567e+01;
	rates[GLU][ILE]=1.7418201388328e+00; 	rates[GLU][LEU]=1.0000000000000e+00;
	rates[GLU][LYS]=1.6519126809071e+02; 	rates[GLU][MET]=1.0000000000000e+00;
	rates[GLU][PHE]=1.4067850525292e+00; 	rates[GLU][PRO]=6.7547121641947e+00;
	rates[GLU][SER]=2.8794794140840e+01; 	rates[GLU][THR]=7.8001372062558e+00;
	rates[GLU][TRP]=1.0000000000000e+00; 	rates[GLU][TYR]=6.9067239183061e+00;
	rates[GLU][VAL]=1.1127702362585e+01; 
	
	rates[GLY][HIS]=1.0000000000000e+00; 	rates[GLY][ILE]=3.1466649021550e+00;
	rates[GLY][LEU]=1.2699794194865e+00; 	rates[GLY][LYS]=1.1962111069278e+01;
	rates[GLY][MET]=1.0000000000000e+00; 	rates[GLY][PHE]=1.0000000000000e+00;
	rates[GLY][PRO]=1.0000000000000e+00; 	rates[GLY][SER]=6.6277950574411e+01;
	rates[GLY][THR]=5.8800079133028e+00; 	rates[GLY][TRP]=5.7494182626674e+00;
	rates[GLY][TYR]=1.6887657206208e+00; 	rates[GLY][VAL]=1.3320553471351e+00;

	rates[HIS][ILE]=6.4536986087271e+00; 	rates[HIS][LEU]=6.0472584534958e+00;
	rates[HIS][LYS]=6.7197196398961e+01; 	rates[HIS][MET]=6.2977633277779e+00;
	rates[HIS][PHE]=2.5347805183364e+01; 	rates[HIS][PRO]=3.2089868698728e+01;
	rates[HIS][SER]=4.0766987134407e+01; 	rates[HIS][THR]=2.3570850628539e+01;
	rates[HIS][TRP]=3.7286635325194e+00; 	rates[HIS][TYR]=3.5270764890474e+02;
	rates[HIS][VAL]=1.0000000000000e+00; 
	
	rates[ILE][LEU]=1.7320653206333e+02; 	rates[ILE][LYS]=1.0298655619743e+01;
	rates[ILE][MET]=2.7262244199514e+02; 	rates[ILE][PHE]=4.4561065036310e+01;
	rates[ILE][PRO]=1.0856482766156e+01; 	rates[ILE][SER]=2.5107659603898e+01;
	rates[ILE][THR]=1.9391167162525e+02; 	rates[ILE][TRP]=1.0000000000000e+00;
	rates[ILE][TYR]=1.3161329199391e+01; 	rates[ILE][VAL]=6.4365086389428e+02;

	rates[LEU][LYS]=7.8314019154706e+00; 	rates[LEU][MET]=2.8290920517725e+02;
	rates[LEU][PHE]=1.1371735519833e+02; 	rates[LEU][PRO]=2.1105885757279e+01;
	rates[LEU][SER]=3.8741359395934e+01; 	rates[LEU][THR]=6.6524559321657e+01;
	rates[LEU][TRP]=1.7071378554833e+01; 	rates[LEU][TYR]=2.3234516108847e+01;
	rates[LEU][VAL]=4.8247261078055e+01; 
	
	rates[LYS][MET]=4.8092094826036e+01; 	rates[LYS][PHE]=3.3887559483420e+00;
	rates[LYS][PRO]=2.6368577564199e+01; 	rates[LYS][SER]=5.5679895711418e+01;
	rates[LYS][THR]=7.1750284708933e+01; 	rates[LYS][TRP]=1.2631893872825e+01;
	rates[LYS][TYR]=2.6932728996777e+01; 	rates[LYS][VAL]=1.0000000000000e+00;

	rates[MET][PHE]=4.7798798034572e+01; 	rates[MET][PRO]=9.9165053447429e+00;
	rates[MET][SER]=5.8505442466161e+01; 	rates[MET][THR]=2.7798190504760e+02;
	rates[MET][TRP]=1.1427000119701e+01; 	rates[MET][TYR]=2.1029990530586e+01;
	rates[MET][VAL]=2.0397078683768e+02; 
	
	rates[PHE][PRO]=9.1089574817139e+00; 	rates[PHE][SER]=3.3835737720574e+01;
	rates[PHE][THR]=1.7815549567056e+01; 	rates[PHE][TRP]=4.1272404968214e+00;
	rates[PHE][TYR]=2.4504156395152e+02; 	rates[PHE][VAL]=3.3435675442163e+00;

	rates[PRO][SER]=8.9421193040709e+01; 	rates[PRO][THR]=6.7485067008375e+01;
	rates[PRO][TRP]=2.2161693733113e+00; 	rates[PRO][TYR]=8.5338209390745e+00;
	rates[PRO][VAL]=4.3342126659660e+00; 
	
	rates[SER][THR]=3.1432036618746e+02; 	rates[SER][TRP]=2.0305343047059e+01;
	rates[SER][TYR]=3.4167877957799e+01; 	rates[SER][VAL]=1.0000000000000e+00;

	rates[THR][TRP]=5.2559565123081e+00; 	rates[THR][TYR]=2.0382362288681e+01;
	rates[THR][VAL]=1.0765527137500e+02; 
	
	rates[TRP][TYR]=1.3814733274637e+01; 	rates[TRP][VAL]=2.8259139240676e+00;

	rates[TYR][VAL]=1.0000000000000e+00; 


        // fill lower triangle of Q*
        for(int i=0;i<alf;++i) {
           for(int j=0;j<i;++j)
	      rates[i][j]=rates[j][i];
           rates[i][i]=0.0;  // Q* matrix!
        }
   
	/* amino acid frequencies */
	statefreqs[ALA]=0.072; statefreqs[ARG]=0.019; statefreqs[ASN]=0.039; statefreqs[ASP]=0.019; statefreqs[CYS]=0.006;
	statefreqs[GLN]=0.025; statefreqs[GLU]=0.024; statefreqs[GLY]=0.056; statefreqs[HIS]=0.028; statefreqs[ILE]=0.088;
	statefreqs[LEU]=0.168; statefreqs[LYS]=0.023; statefreqs[MET]=0.054; statefreqs[PHE]=0.061; statefreqs[PRO]=0.054;
	statefreqs[SER]=0.072; statefreqs[THR]=0.086; statefreqs[TRP]=0.029; statefreqs[TYR]=0.033; statefreqs[VAL]=0.043;
}

void substmodel::blosum62data()
{
	/*
	 * BLOSUM62 model of amino acid evolution
	 * 
	 * S. Henikoff and J. G. Henikoff.  1992.  PNAS USA 89:10915-10919.
	 * 
	 */

	rates[ALA][ARG]=7.3579038969751e-01; 	rates[ALA][ASN]=4.8539105546575e-01;
	rates[ALA][ASP]=5.4316182089867e-01; 	rates[ALA][CYS]=1.4599953104700e+00;
	rates[ALA][GLN]=1.1997057046020e+00; 	rates[ALA][GLU]=1.1709490427999e+00;
	rates[ALA][GLY]=1.9558835749595e+00; 	rates[ALA][HIS]=7.1624144499779e-01;
	rates[ALA][ILE]=6.0589900368677e-01; 	rates[ALA][LEU]=8.0001653051838e-01;
	rates[ALA][LYS]=1.2952012667833e+00; 	rates[ALA][MET]=1.2537582666635e+00;
	rates[ALA][PHE]=4.9296467974759e-01; 	rates[ALA][PRO]=1.1732759009239e+00;
	rates[ALA][SER]=4.3250926870566e+00; 	rates[ALA][THR]=1.7291780194850e+00;
	rates[ALA][TRP]=4.6583936772479e-01; 	rates[ALA][TYR]=7.1820669758623e-01;
	rates[ALA][VAL]=2.1877745220045e+00; 
	
	rates[ARG][ASN]=1.2974467051337e+00; 	rates[ARG][ASP]=5.0096440855513e-01;
	rates[ARG][CYS]=2.2782657420895e-01; 	rates[ARG][GLN]=3.0208336100636e+00;
	rates[ARG][GLU]=1.3605741904203e+00; 	rates[ARG][GLY]=4.1876330851753e-01;
	rates[ARG][HIS]=1.4561411663360e+00; 	rates[ARG][ILE]=2.3203644514174e-01;
	rates[ARG][LEU]=6.2271166969249e-01; 	rates[ARG][LYS]=5.4111151414889e+00;
	rates[ARG][MET]=9.8369298745695e-01; 	rates[ARG][PHE]=3.7164469320875e-01;
	rates[ARG][PRO]=4.4813366171831e-01; 	rates[ARG][SER]=1.1227831042096e+00;
	rates[ARG][THR]=9.1466595456337e-01; 	rates[ARG][TRP]=4.2638231012175e-01;
	rates[ARG][TYR]=7.2051744121611e-01; 	rates[ARG][VAL]=4.3838834377202e-01;

	rates[ASN][ASP]=3.1801000482161e+00; 	rates[ASN][CYS]=3.9735894989702e-01;
	rates[ASN][GLN]=1.8392161469920e+00; 	rates[ASN][GLU]=1.2404885086396e+00;
	rates[ASN][GLY]=1.3558723444845e+00; 	rates[ASN][HIS]=2.4145014342081e+00;
	rates[ASN][ILE]=2.8301732627800e-01; 	rates[ASN][LEU]=2.1188815961519e-01;
	rates[ASN][LYS]=1.5931370434574e+00; 	rates[ASN][MET]=6.4844127878707e-01;
	rates[ASN][PHE]=3.5486124922252e-01; 	rates[ASN][PRO]=4.9488704370192e-01;
	rates[ASN][SER]=2.9041016564560e+00; 	rates[ASN][THR]=1.8981736345332e+00;
	rates[ASN][TRP]=1.9148204624678e-01; 	rates[ASN][TYR]=5.3822251903674e-01;
	rates[ASN][VAL]=3.1285879799342e-01; 
	
	rates[ASP][CYS]=2.4083661480204e-01; 	rates[ASP][GLN]=1.1909457033960e+00;
	rates[ASP][GLU]=3.7616252083685e+00; 	rates[ASP][GLY]=7.9847324896839e-01;
	rates[ASP][HIS]=7.7814266402188e-01; 	rates[ASP][ILE]=4.1855573246161e-01;
	rates[ASP][LEU]=2.1813157759360e-01; 	rates[ASP][LYS]=1.0324479249521e+00;
	rates[ASP][MET]=2.2262189795786e-01; 	rates[ASP][PHE]=2.8173069420651e-01;
	rates[ASP][PRO]=7.3062827299842e-01; 	rates[ASP][SER]=1.5827541420653e+00;
	rates[ASP][THR]=9.3418750943056e-01; 	rates[ASP][TRP]=1.4534504627853e-01;
	rates[ASP][TYR]=2.6142220896504e-01; 	rates[ASP][VAL]=2.5812928941763e-01;

	rates[CYS][GLN]=3.2980150463028e-01; 	rates[CYS][GLU]=1.4074889181440e-01;
	rates[CYS][GLY]=4.1820319228376e-01; 	rates[CYS][HIS]=3.5405810983129e-01;
	rates[CYS][ILE]=7.7489402279418e-01; 	rates[CYS][LEU]=8.3184264014158e-01;
	rates[CYS][LYS]=2.8507880090648e-01; 	rates[CYS][MET]=7.6768882347954e-01;
	rates[CYS][PHE]=4.4133747118660e-01; 	rates[CYS][PRO]=3.5600849876863e-01;
	rates[CYS][SER]=1.1971884150942e+00; 	rates[CYS][THR]=1.1198313585160e+00;
	rates[CYS][TRP]=5.2766441887169e-01; 	rates[CYS][TYR]=4.7023773369610e-01;
	rates[CYS][VAL]=1.1163524786062e+00; 
	
	rates[GLN][GLU]=5.5289191779282e+00; 	rates[GLN][GLY]=6.0984630538281e-01;
	rates[GLN][HIS]=2.4353411311401e+00; 	rates[GLN][ILE]=2.3620245120365e-01;
	rates[GLN][LEU]=5.8073709318144e-01; 	rates[GLN][LYS]=3.9452776745146e+00;
	rates[GLN][MET]=2.4948960771127e+00; 	rates[GLN][PHE]=1.4435695975031e-01;
	rates[GLN][PRO]=8.5857057567418e-01; 	rates[GLN][SER]=1.9348709245965e+00;
	rates[GLN][THR]=1.2774802945956e+00; 	rates[GLN][TRP]=7.5865380864172e-01;
	rates[GLN][TYR]=9.5898974285014e-01; 	rates[GLN][VAL]=5.3078579012486e-01;

	rates[GLU][GLY]=4.2357999217628e-01; 	rates[GLU][HIS]=1.6268910569817e+00;
	rates[GLU][ILE]=1.8684804693170e-01; 	rates[GLU][LEU]=3.7262517508685e-01;
	rates[GLU][LYS]=2.8024271516787e+00; 	rates[GLU][MET]=5.5541539747043e-01;
	rates[GLU][PHE]=2.9140908416530e-01; 	rates[GLU][PRO]=9.2656393484598e-01;
	rates[GLU][SER]=1.7698932389373e+00; 	rates[GLU][THR]=1.0710972360073e+00;
	rates[GLU][TRP]=4.0763564893830e-01; 	rates[GLU][TYR]=5.9671930034577e-01;
	rates[GLU][VAL]=5.2425384633796e-01; 
	
	rates[GLY][HIS]=5.3985912495418e-01; 	rates[GLY][ILE]=1.8929629237636e-01;
	rates[GLY][LEU]=2.1772115923623e-01; 	rates[GLY][LYS]=7.5204244030271e-01;
	rates[GLY][MET]=4.5943617357855e-01; 	rates[GLY][PHE]=3.6816646445253e-01;
	rates[GLY][PRO]=5.0408659952683e-01; 	rates[GLY][SER]=1.5093262532236e+00;
	rates[GLY][THR]=6.4143601140497e-01; 	rates[GLY][TRP]=5.0835892463812e-01;
	rates[GLY][TYR]=3.0805573703500e-01; 	rates[GLY][VAL]=2.5334079019018e-01;

	rates[HIS][ILE]=2.5271844788492e-01; 	rates[HIS][LEU]=3.4807220979697e-01;
	rates[HIS][LYS]=1.0225070358890e+00; 	rates[HIS][MET]=9.8431152535870e-01;
	rates[HIS][PHE]=7.1453370392764e-01; 	rates[HIS][PRO]=5.2700733915060e-01;
	rates[HIS][SER]=1.1170297629105e+00; 	rates[HIS][THR]=5.8540709022472e-01;
	rates[HIS][TRP]=3.0124860078016e-01; 	rates[HIS][TYR]=4.2189539693890e+00;
	rates[HIS][VAL]=2.0155597175031e-01; 
	
	rates[ILE][LEU]=3.8909637733035e+00; 	rates[ILE][LYS]=4.0619358664202e-01;
	rates[ILE][MET]=3.3647977631042e+00; 	rates[ILE][PHE]=1.5173593259539e+00;
	rates[ILE][PRO]=3.8835540920564e-01; 	rates[ILE][SER]=3.5754441245967e-01;
	rates[ILE][THR]=1.1790911972601e+00; 	rates[ILE][TRP]=3.4198578754023e-01;
	rates[ILE][TYR]=6.7461709322842e-01; 	rates[ILE][VAL]=8.3118394054582e+00;

	rates[LEU][LYS]=4.4557027426059e-01; 	rates[LEU][MET]=6.0305593795716e+00;
	rates[LEU][PHE]=2.0648397032375e+00; 	rates[LEU][PRO]=3.7455568747097e-01;
	rates[LEU][SER]=3.5296918452729e-01; 	rates[LEU][THR]=9.1525985769421e-01;
	rates[LEU][TRP]=6.9147463459998e-01; 	rates[LEU][TYR]=8.1124585632307e-01;
	rates[LEU][VAL]=2.2314056889131e+00; 
	
	rates[LYS][MET]=1.0730611843319e+00; 	rates[LYS][PHE]=2.6692475051102e-01;
	rates[LYS][PRO]=1.0473834507215e+00; 	rates[LYS][SER]=1.7521659178195e+00;
	rates[LYS][THR]=1.3038752007987e+00; 	rates[LYS][TRP]=3.3224304063396e-01;
	rates[LYS][TYR]=7.1799348690032e-01; 	rates[LYS][VAL]=4.9813847530407e-01;

	rates[MET][PHE]=1.7738551688305e+00; 	rates[MET][PRO]=4.5412362510273e-01;
	rates[MET][SER]=9.1872341574605e-01; 	rates[MET][THR]=1.4885480537218e+00;
	rates[MET][TRP]=8.8810109815193e-01; 	rates[MET][TYR]=9.5168216224591e-01;
	rates[MET][VAL]=2.5758507553153e+00; 
	
	rates[PHE][PRO]=2.3359790962888e-01; 	rates[PHE][SER]=5.4002764482413e-01;
	rates[PHE][THR]=4.8820611879305e-01; 	rates[PHE][TRP]=2.0743248934965e+00;
	rates[PHE][TYR]=6.7472604308008e+00; 	rates[PHE][VAL]=8.3811961017754e-01;

	rates[PRO][SER]=1.1691295777157e+00; 	rates[PRO][THR]=1.0054516831488e+00;
	rates[PRO][TRP]=2.5221483002727e-01; 	rates[PRO][TYR]=3.6940531935451e-01;
	rates[PRO][VAL]=4.9690841067567e-01; 
	
	rates[SER][THR]=5.1515562922704e+00; 	rates[SER][TRP]=3.8792562209837e-01;
	rates[SER][TYR]=7.9675152076106e-01; 	rates[SER][VAL]=5.6192545744165e-01;

	rates[THR][TRP]=5.1312812689059e-01; 	rates[THR][TYR]=8.0101024319939e-01;
	rates[THR][VAL]=2.2530740511763e+00; 
	
	rates[TRP][TYR]=4.0544190065580e+00; 	rates[TRP][VAL]=2.6650873142646e-01;

	rates[TYR][VAL]=1.0000000000000e+00; 

        // fill lower triangle of Q*
        for(int i=0;i<alf;++i) {
          for(int j=0;j<i;++j)
	      rates[i][j]=rates[j][i];
          rates[i][i]=0.0;  // Q* matrix!
        }
   
	statefreqs[ALA]=0.074; statefreqs[ARG]=0.052; statefreqs[ASN]=0.045; statefreqs[ASP]=0.054;
	statefreqs[CYS]=0.025; statefreqs[GLN]=0.034; statefreqs[GLU]=0.054; statefreqs[GLY]=0.074;
	statefreqs[HIS]=0.026; statefreqs[ILE]=0.068; statefreqs[LEU]=0.099; statefreqs[LYS]=0.058;
	statefreqs[MET]=0.025; statefreqs[PHE]=0.047; statefreqs[PRO]=0.039; statefreqs[SER]=0.057;
	statefreqs[THR]=0.051; statefreqs[TRP]=0.013; statefreqs[TYR]=0.032; statefreqs[VAL]=0.073;
}



/* guess AA model of substitution */
void substmodel::guessmodel(vector<double>& empfreqs)
{
	double c1, c2, c3, c4;
	int i;

	
	/* chi2 fit to amino acid frequencies */
		
	/* chi2 distance Dayhoff */
	dyhfdata();
	c1 = 0;
	for (i = 0; i < 20; ++i)
		c1 = c1 + (empfreqs[i]-statefreqs[i])*(empfreqs[i]-statefreqs[i]);
	/* chi2 distance JTT */
	jttdata();
	c2 = 0;
	for (i = 0; i < 20; ++i)
		c2 = c2 + (empfreqs[i]-statefreqs[i])*(empfreqs[i]-statefreqs[i]);
	/* chi2 distance mtREV */
	mtrevdata();
	c3 = 0;
	for (i = 0; i < 20; ++i)
		c3 = c3 + (empfreqs[i]-statefreqs[i])*(empfreqs[i]-statefreqs[i]);
	/* chi2 distance Blosum62 */
	blosum62data();
	c4 = 0;
	for (i = 0; i < 20; ++i)
		c4 = c4 + (empfreqs[i]-statefreqs[i])*(empfreqs[i]-statefreqs[i]);
	
	if(c1 < c2) {
	     if(c1 < c3) {
	          if(c1 < c4) {
		     cout << "Dayhoff model fitted best\n";
		     model = DAYHOFF;
		  } else {
		     cout << "Blosum62 model fitted best\n";
		     model = BLOSUM62;
		  }
	     } else {
	          if(c3 < c4) {
		     cout << "mtRev24 model fitted best\n";
		     model = MTREV24;
		  } else {
		     cout << "Blosum62 model fitted best\n";
		     model = BLOSUM62;
		  }
	     }
	} else {
	     if(c2 < c3) {
	          if(c2 < c4) {
		     cout << "JTT model fitted best\n";
		     model = JTT;
		  } else {
		     cout << "Blosum62 model fitted best\n";
		     model = BLOSUM62;
		  }
	     } else {
	          if(c3 < c4) {
		     cout << "mtRev24 model fitted best\n";
		     model = MTREV24;
		  } else {
		     cout << "Blosum62 model fitted best\n";
		     model = BLOSUM62;
		  }
	     }
	}

}

void substmodel::genrevmod()
{
    int i, j, k;
    if(params.size()!=6) {
       cerr << "The general reversible model needs six parameters,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
    }
   
    k=0;
    for(i=0;i<alf;++i)
      for(j=i+1;j<alf;++j) {
         rates[i][j]=params[k];
	 ++k;
    }
    // fill lower triangle of Q*
    for(i=0;i<alf;++i) {
       for(j=0;j<i;++j)
	   rates[i][j]=rates[j][i];
       rates[i][i]=0.0;  // Q* matrix!
    }
}

void substmodel::tamneimod()
{
   double kappa, tau, alp, alpr, alpy;
   if(params.size()!=2) {
       cerr << "The Tamura Nei model needs two parameters,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
   }
  
   tau = params[0];
   kappa = params[1];
   alp = 2.0*tau;
   alpr = (alp * 2.0) / (kappa + 1.0);
   alpy = kappa * alpr;

   // Q*			
   rates[dAtp][dAtp] = 0.0;                rates[dAtp][dCtp] = 1.0;                rates[dAtp][dGtp] = alpr;               rates[dAtp][dTtp] = 1.0;
   rates[dCtp][dAtp] = rates[dAtp][dCtp];  rates[dCtp][dCtp] = 0.0;                rates[dCtp][dGtp] = 1.0;                rates[dCtp][dTtp] = alpy;
   rates[dGtp][dAtp] = rates[dAtp][dGtp];  rates[dGtp][dCtp] = rates[dCtp][dGtp];  rates[dGtp][dGtp] = 0.0;                rates[dGtp][dTtp] = 1.0;
   rates[dTtp][dAtp] = rates[dAtp][dTtp];  rates[dTtp][dCtp] = rates[dCtp][dTtp];  rates[dTtp][dGtp] = rates[dGtp][dTtp];  rates[dTtp][dTtp] = 0.0; 

}

void substmodel::hkymod()
{
   double tau;
   if(params.size()!=1) {
       cerr << "The HKY model needs one parameter,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
   }
  
   tau = 2.0*params[0];
   // Q*			
   rates[dAtp][dAtp] = 0.0;                rates[dAtp][dCtp] = 1.0;                rates[dAtp][dGtp] = tau;                rates[dAtp][dTtp] = 1.0;
   rates[dCtp][dAtp] = rates[dAtp][dCtp];  rates[dCtp][dCtp] = 0.0;                rates[dCtp][dGtp] = 1.0;                rates[dCtp][dTtp] = tau;
   rates[dGtp][dAtp] = rates[dAtp][dGtp];  rates[dGtp][dCtp] = rates[dCtp][dGtp];  rates[dGtp][dGtp] = 0.0;                rates[dGtp][dTtp] = 1.0;
   rates[dTtp][dAtp] = rates[dAtp][dTtp];  rates[dTtp][dCtp] = rates[dCtp][dTtp];  rates[dTtp][dGtp] = rates[dGtp][dTtp];  rates[dTtp][dTtp] = 0.0; 
}

void substmodel::kimura2mod()
{
   double tau;
   if(params.size()!=1) {
       cerr << "The Kimura 2 parameter model needs one parameter,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
   }
  
   tau = 2.0*params[0];
   // Q*			
   rates[dAtp][dAtp] = 0.0;                rates[dAtp][dCtp] = 1.0;                rates[dAtp][dGtp] = tau;                rates[dAtp][dTtp] = 1.0;
   rates[dCtp][dAtp] = rates[dAtp][dCtp];  rates[dCtp][dCtp] = 0.0;                rates[dCtp][dGtp] = 1.0;                rates[dCtp][dTtp] = tau;
   rates[dGtp][dAtp] = rates[dAtp][dGtp];  rates[dGtp][dCtp] = rates[dCtp][dGtp];  rates[dGtp][dGtp] = 0.0;                rates[dGtp][dTtp] = 1.0;
   rates[dTtp][dAtp] = rates[dAtp][dTtp];  rates[dTtp][dCtp] = rates[dCtp][dTtp];  rates[dTtp][dGtp] = rates[dGtp][dTtp];  rates[dTtp][dTtp] = 0.0; 

   statefreqs[0]=statefreqs[1]=statefreqs[2]=statefreqs[3]=0.25;
}

void substmodel::kimura3mod()
{

   double kappa, tau, alp, alpr, alpy;
   if(params.size()!=2) {
       cerr << "The Kimura 3 parameter model needs two parameters,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
   }
  
   tau = params[0];
   kappa = params[1];
   alp = 2.0*tau;
   alpr = (alp * 2.0) / (kappa + 1.0);
   alpy = kappa * alpr;

   // Q*			
   rates[dAtp][dAtp] = 0.0;                rates[dAtp][dCtp] = 1.0;                rates[dAtp][dGtp] = alpr;               rates[dAtp][dTtp] = 1.0;
   rates[dCtp][dAtp] = rates[dAtp][dCtp];  rates[dCtp][dCtp] = 0.0;                rates[dCtp][dGtp] = 1.0;                rates[dCtp][dTtp] = alpy;
   rates[dGtp][dAtp] = rates[dAtp][dGtp];  rates[dGtp][dCtp] = rates[dCtp][dGtp];  rates[dGtp][dGtp] = 0.0;                rates[dGtp][dTtp] = 1.0;
   rates[dTtp][dAtp] = rates[dAtp][dTtp];  rates[dTtp][dCtp] = rates[dCtp][dTtp];  rates[dTtp][dGtp] = rates[dGtp][dTtp];  rates[dTtp][dTtp] = 0.0; 

   statefreqs[0]=statefreqs[1]=statefreqs[2]=statefreqs[3]=0.25;
}

void substmodel::fel81mod()
{
   int i, j;

   // Q*			
   for(i=0;i<alf;++i) {
     for(j=i+1;j<alf;++j)
        rates[i][j]=1.0;
     // fill lower triangle of Q*
     for(j=0;j<i;++j)
	rates[i][j]=1.0;
     rates[i][i]=0.0;
   }
}

void substmodel::jcmod()
{
   int i, j;

   // Q*			
   for(i=0;i<alf;++i) {
     for(j=i+1;j<alf;++j)
        rates[i][j]=1.0;
     // fill lower triangle of Q*
     for(j=0;j<i;++j)
	rates[i][j]=1.0;
     rates[i][i]=0.0;
   }

   statefreqs[0]=statefreqs[1]=statefreqs[2]=statefreqs[3]=0.25;
}

void substmodel::mv2000mod()
{
    int i, j, k;
    
    if(params.size()!=190) {
       cerr << "The general reversible model for amino acids needs 190 parameters,\n"
            << "yet there are " << params.size() << endl;
       exit(EXIT_FAILURE);
    }
   
    k=0;
    for(i=0;i<alf;++i)
      for(j=i+1;j<alf;++j) {
         rates[i][j]=params[k];
	 ++k;
    }
    // fill lower triangle of Q*
    for(i=0;i<alf;++i) {
      for(j=0;j<i;++j)
	 rates[i][j]=rates[j][i];
      rates[i][i]=0.0;
    }
}

substmodel substmodel::operator=(substmodel s2)
{
  params=s2.params;
  model=s2.model;
  alf=s2.alf;   
  logalf=s2.logalf;   
  statefreqs=s2.statefreqs;
  //logstatefreqs=s2.logstatefreqs;
  rates=s2.rates;
  time=s2.time;
  probs_time=s2.probs_time;
  //logprobs_time=s2.logprobs_time;
  Eval=s2.Eval;
  Evec=s2.Evec;
  Ievc=s2.Ievc;
  iexp=s2.iexp;

  return *this;
}

alphsize substmodel::retalf()
{
  return alf;
}


int substmodel::choosebase()
{
   int i;
   double R=randomunitintervall();
   double probsum=0.0;
   for(i=0;i<alf;++i) {
       probsum+=statefreqs[i];
       if(R<probsum)
          return i;
   }
   cerr << "Reached the end of function choosebase\n";
   exit(EXIT_FAILURE);
   
   return -1;
}


bool substmodel::choosepair(int& x, int& y) // gibt zufaelliges Paar zurueck (gezogen aus statefreqs*probs_time)
{
   int i, j;
   double R=randomunitintervall();
   double probsum=0.0;

   for(i=0;i<alf;++i)
     for(j=0;j<alf;++j) {
         probsum+=statefreqs[i]*probs_time[i][j];
         if(R<probsum) {
	    x=i;
	    y=j;
	    return TRUE;
	 }
     }
        
   cerr << "Reached the end of function choosepair\n";
   exit(EXIT_FAILURE);
   
   return FALSE;
}


int substmodel::evolbase(int profpar)
{
   int i;
   double logprobsum, probsum2=0.0;
   vector<double> logliks(alf);
   double R=randomunitintervall();
   
   // Wenn die Aeste sehr kurz sind, wird die gleiche Base genommen
   if(time<INFACC*DBL_MIN)
      return profpar;

   for(i=0;i<alf;++i)
       logliks[i]=logfreqofstate(profpar)+logProb_ij(profpar,i);
   logprobsum=untidy2_addlog(logliks);
   for(i=0;i<alf;++i) {
       probsum2+=exp(logliks[i]-logprobsum);
       if(R<=probsum2)
          return i;
   }
   cerr << "Warning: Reached end of evolbase\n";
   return profpar;
}


