/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#ifndef SUBSTMOD_H
#define SUBSTMOD_H

#define FALSE 0
#define TRUE 1


class substmodel {
private :
  vector<double> params;
  subarch model;
  alphsize alf;
  int logalf;  // Groesse von logstatefreqs und logprobs_time
  vector<double> statefreqs;
  vector<vector<double> > rates; // 1 Pam
  double time;
  vector<vector<double> > probs_time;

  void tranprobmat();
  void tprobmtrx();
  void onepamratematrix();
  void fillogprobs();
  void fillogstatefreqs();
  
  // AA matrices (Q*)
  void mv2000mod();
  void jttdata();
  void dyhfdata();
  void mtrevdata();
  void blosum62data();
  void guessmodel(vector<double>& empfreqs);
  
  // Nuc matrices (Q*)
  void genrevmod();
  void tamneimod();
  void hkymod();
  void kimura2mod();
  void kimura3mod();
  void fel81mod();
  void jcmod();

  // eigen system
  vector<double> Eval;
  vector<vector<double> > Evec, Ievc, iexp;
  void eigensystem();
  void luinverse();
  void checkevector();
  /***************************** internal functions *****************************/
  void elmhes(double **b, int *ordr, int n);
  void eltran(double **b, int *ordr, int n);
  void mcdiv(double ar, double ai, double br, double bi, double *cr, double *ci);
  void hqr2(int n, int low, int hgh, double **b, double *wi);
  
public :
  substmodel();  // default constructor => JC69
  substmodel(alphsize a);
  substmodel(alphsize a, subarch m);
  substmodel(alphsize a, subarch m, vector<double>& freqs);
  substmodel(alphsize a, subarch m, vector<double>& freqs, vector<double>& parameter);
  substmodel(const substmodel& s);
  ~substmodel();
  double logfreqofstate(int state);
  double basefreq(char base);
  double amacfreq(char amac);
  double logProb_ij(int i, int j);    // ohne Ueberpruefung der Zeit
  double logProb_ij(int i, int j, double time);
  // double Prob_ij(int i, int j);    // ohne Ueberpruefung der Zeit
  // double Prob_ij(int i, int j, double time);
  // double Prob_cd(char c, char d);    // ohne Ueberpruefung der Zeit
  // double Prob_cd(char c, char d, double time);
  // double NucProb_cd(char c, char d);
  // double AAProb_cd(char c, char d);
  // Gewichtung der Zustandshaeufigkeiten und Ue'w'keiten mit Profilen 
  // double Erw_freqofstate(const vector<double>& prof);
  // double Erw_Prob_prof(const vector<double>& prof1, const vector<double>& prof2);// ohne Ueberpruefung der Zeit
  // double Erw_Prob_prof(const vector<double>& prof1, const vector<double>& prof2, double time);
  
  void settime(double ptime);
  void setfreqs(vector<double>& freqs);
  void setparams(vector<double>& p);
  double timeOf();
  void freqsOf(vector<double>& f);
  alphsize retalf();
  subarch paramsOf(vector<double>& p);
  int choosebase(); // gibt zufaellige Base zurueck (gezogen aus statefreqs)
  bool choosepair(int& x, int& y); // gibt zufaelliges Paar zurueck (gezogen aus statefreqs*probs_time)
  int evolbase(int profpar);
  substmodel operator=(substmodel s2);
};

#endif
