/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#ifndef TREESPEC_H
#define TREESPEC_H

#define FALSE 0
#define TRUE 1
#define DEFAULTBRANCH 0.1 // willkuerliche Astlaenge

class Newick {
public:
   Newick(string v);
   void desc_lab_len(string v);
   void subtrees();
   string desc_list;
   double len_to_par;
   bool leaf;
   //string links;
   //string rechts;
   vector<string> Kinder;
};

class clustal : public string {  // Newick-kodierter Baum ohne Kommentare
public:
   clustal(const char *filename);
   clustal(ifstream& datei);
};


class identifier {  // alles, was einen Knoten im Baum identifiziert
public :   
   string name;  // Name der Sequenz
   string treefromhere;  // Newickkodierung mit aktuellem Knoten als Wurzel
   int number;  // Nummer der Sequenz im Datensatz;
   void setnumber(vector<vector<char> > Namen); // finde zum Sequenznamen passende Nummer
   identifier operator= (identifier id2);  // Zuweisungsoperator
   bool operator== (identifier& id2);  // Vergleichsoperator
};

class baum;

class neighbor {
public :
   double length;  // Abstand zum Nachbarknoten
   bool fix;       // Soll der Ast zum Nachbarn festgehalten werden?
   bool toroot;    // Fuehrt der Weg zur Wurzel ueber den Nachbarn?
   baum *nachbar;  // Nachbarknoten
   vector<int> ausrichter;  // Alignment zur Sequenz des Nachbarn

   neighbor();
   neighbor operator= (neighbor nb2);  // Zuweisungsoperator
   neighbor operator<= (neighbor nb2);  // Zuweisungsoperator ohne Zuweisung vom ausrichter
//   bool operator== (const neighbor& nb2);  // Vergleichsoperator
//   bool operator!= (const neighbor& nb2);  // Vergleichsoperator
};

class bioseq : public vector<int> {
public:
   bioseq();
   bioseq(vector<char>& seq, int typ);
  ~bioseq();
   int laen() const;
};



class baum {
public :
   identifier id;   // alles, was den Knoten identifiziert
   vector<neighbor> nb;  // Verweise auf die Nachbarn
   bioseq pro;  // Sequenz
   
   baum(identifier ida); // erzeugt ein Blatt mit Wert ida
   baum(baum& source); // copy constructor
   baum(baum& source, bool topologyonly);  // kopiert nur die Topologie (d.h. ohne Sequenzen und ausrichter)
   baum(string s, bool anf);  // erzeugt Baum aus Newick-Code
   ~baum();

   void NeighborJoining(vector<vector<double> >& distmat, unsigned int& taxcount);
     
   // Constant member functions
   size_t depth() const;  // Entfernung von der Wurzel
   size_t height() const;
   size_t size() const;
   size_t numleaves() const;
   double laenge() const;
   double aprxspan() const;
   double aprxheight() const;
   double opt_delta() const;
   bool isLeaf() const;
   bool isRoot() const;
   baum *NB(int i) const; // i-ter Nachbar
   int  whichone(const baum *T) const; // Index von Nachbarknoten T
   baum *parent() const; // Nachbar in Richtung der Wurzel
   baum *root(); // Wurzel
   baum *findLeaf(identifier& id2); // findet Blatt mit Wert id2

   // Modification member functions
   void setid(const identifier& ida);  // setzt neuen Identifier 
   void swapbr(int nbkn, int ownch, int nbch); // tauscht Kind nbch von nbkn mit eigenem Kind ownch (Vorsicht mit Wurzel!!!!)
   double rndbrl(double oldl, double delta, bool fix);
   double shiftnodes(double delta, int idx); //  verschiebt zwei Nachbarn (this und nb[idx].nachbar) ,
                                             // gibt neue Astlaenge zurueck
   double shiftnodes(double delta, int idx, bool allfix); //  verschiebt zwei Nachbarn (this und nb[idx].nachbar) ,
                                                          // gibt neue Astlaenge zurueck
   baum *pickrandleaf();  // waehlt zufaellig ein Blatt
   baum *reroot();  // macht eins der Blaetter zur Wurzel
   void reroot(baum *newp);  // macht Wurzel zur Wurzel
   void bifurcate(bool anf);// erzeugt vollstaendig bifurkierenden Baum ohne ueberfluessige Knoten
   void fuseblindnodes(bool anf); // verhindert im ganzen Baum *-*-* (wie bifurcate ohne Aufruf von splitnode)
   void splitnode(); // loest anfaengliche Multifurkationen auf
   void fusenodes(); // verhindert  *--*--*
   void allebisauf(vector<int> &v, unsigned int size, int except);
   void update_tree(); // nur Teilbaum, falls man nicht mit der Wurzel beginnt
   string clustname(); // string mit Nummern der Blaetter im Cluster
   void edges(vector<baum *>& nodes, vector<unsigned int>& nbidx); // erfasst alle Kanten im Baum
   void internaledges(vector<baum *>& nodes, vector<unsigned int>& nbidx); // erfasst alle inneren Kanten im Baum
   void intnodes(vector<baum *>& nodes); // erfasst alle inneren Knoten im Baum
   int rndchild(int ausser); // zufaelliger Nachbar!=nb[ausser]
   void print(ostream& trstr);
   baum* teilbaum(vector<string>& IDs, vector<char>& pattern);// erzeugt Teilbaum, der Taxa, die im pattern
   baum* abuse_fix(vector<string>& IDs, vector<char>& pattern);
   void findnames(baum*& p, vector<bool>& nucpres, vector<string>& IDs, vector<bool>& namefound);
   bool post_order_abuse_fix(vector<bool>& nucpres);
   void pre_order_abuse_fix();
   void trim();

   bool evolve(substmodel& sub, indel& idmod); // Datensatz entlang des Baumes evolvieren lassen
   void clear(); // entfernt pro und ausrichter

   bool operator== (baum& b2);  // Vergleichsoperator fuer die Topologie
};

#endif
